"""
Barlines can appear at mid-measure positions, without using an implicit measure!
"""
from musicxml import XMLBarline

"""
MuseScore supports mid-measure barlines
"""

from pathlib import Path

from musicscore import Score, Chord
from musicscore.tests.util import IdTestCase


class TestLily46b(IdTestCase):
    def test_lily_46b_MidmeasureBarlines(self):
        score = Score()
        p = score.add_part('p1')
        midis = [72, 69, 65, 72]
        chords = [Chord(m, 1) for m in midis]
        [p.add_chord(ch) for ch in chords]
        b = XMLBarline(location='middle')
        b.xml_bar_style = 'dashed'
        chords[1].add_after_note_xml_objects(b)
        score.finalize()

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from pathlib import Path

from musicscore import Score, G, A, B, C, D, E, F, Chord, Midi, Accidental
from musicscore.tests.util import IdTestCase


class TestAccidentals(IdTestCase):
    def test_modes_in_quarter_steps(self):

        """
        All pitches from c4 to c5 in ascending quarter steps; different Accidental.mode: standard, enharmonic, sharp,
        flat, force-sharp, force-flat.
        """

        def create_part():
            scale = [Midi(m / 2) for m in range(120, 144)]
            p = score.add_part(f'{mode}')
            p.name = f'{mode}'
            for m in scale:
                m.accidental.mode = mode
                ch = Chord(m, 1)
                p.add_chord(ch)

        score = Score('Accidental modes')
        for mode in ['standard', 'enharmonic', 'sharp', 'flat', 'force-sharp', 'force-flat']:
            create_part()
        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from pathlib import Path

from musicscore import Score, G, A, B, C, D, E, F, Chord, Midi
from musicscore.tests.util import IdTestCase

"""
All pitches from G to c”” in ascending steps; First without accidentals, then with a sharp and then with a flat 
accidental. Double alterations and cautionary accidentals are tested at the end.
"""


class TestLily01a(IdTestCase):
    def test_lily_01a_Pitches_Pitches(self):

        score = Score('Pitches and accidentals')
        p = score.add_part('pitches')
        scale_1 = [G(2), A(2), B(2), C(3), D(3), E(3), F(3), G(3), A(3), B(3), C(4), D(4), E(4), F(4), G(4), A(4), B(4),
                   C(5),
                   D(5), E(5), F(5), G(5), A(5), B(5), C(6), D(6), E(6), F(6), G(6), A(6), B(6), C(7)]
        # with sharps
        scale_2 = [midi.__deepcopy__() for midi in scale_1]
        for midi in scale_2:
            midi.transpose(1)
            if midi.value % 12 in [0, 5]:
                # B# and E#
                midi.accidental.mode = 'force-sharp'
            else:
                midi.accidental.mode = 'sharp'
        # with flats
        scale_3 = [midi.__deepcopy__() for midi in scale_1]
        for midi in scale_3:
            midi.transpose(-1)
            if midi.value % 12 in [11, 4]:
                # C-flat and F-flat
                midi.accidental.mode = 'force-flat'
            else:
                midi.accidental.mode = 'flat'
        # Double alterations and cautionary accidentals.
        cs = [Midi(74), Midi(70), Midi(73), Midi(73), Midi(73), Midi(73)]
        cs[0].accidental.mode = 'force-sharp'
        cs[1].accidental.mode = 'force-flat'
        for midi in scale_1 + scale_2 + scale_3 + cs:
            p.add_chord(Chord(midis=midi, quarter_duration=1))
        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



"""
Nested repeats, each with alternative endings.
"""

from pathlib import Path

from musicscore import Score, Chord
from musicscore.tests.util import IdTestCase


class TestLily45d(IdTestCase):
    def test_lily_45d_Repeats_Nested_Alternatives(self):
        score = Score()
        p = score.add_part('p1')
        [p.add_chord(Chord(0, 4)) for _ in range(11)]

        score.set_multi_measure_rest(3, 5)
        score.set_multi_measure_rest(6, 8)

        p.get_measure(2).set_repeat_barline()
        p.get_measure(2).set_repeat_ending(number=1, type='start')
        p.get_measure(2).set_repeat_ending(number=1, type='stop')

        p.get_measure(3).set_repeat_ending(number=2, type='start')
        p.get_measure(5).set_repeat_ending(number=2, type='discontinue')

        p.get_measure(9).set_repeat_barline()
        p.get_measure(9).set_repeat_ending(number=1, type='start')
        p.get_measure(9).set_repeat_ending(number=1, type='stop')

        p.get_measure(10).set_repeat_ending(number=2, type='start')
        p.get_measure(10).set_repeat_ending(number=5, type='discontinue')

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from typing import List

from math import trunc

from musicscore.quantize import QuantizeMixin
from musicxml.xmlelement.xmlelement import XMLNotations, XMLTuplet, XMLTimeModification, XMLBeam
from quicktions import Fraction

from musicscore.chord import _split_copy, _group_chords, Chord
from musicscore.musictree import MusicTree
from musicscore.exceptions import BeatWrongDurationError, BeatIsFullError, BeatHasNoParentError, \
    ChordHasNoQuarterDurationError, \
    ChordHasNoMidisError, AlreadyFinalizedError, BeatNotFullError, AddChordError
from musicscore.finalize import FinalizeMixin
from musicscore.quarterduration import QuarterDuration, QuarterDurationMixin
from musicscore.util import lcm

__all__ = ['SPLITTABLES', 'Beat', '_beam_chord_group']

SPLITTABLES = {
    QuarterDuration(0): {
        QuarterDuration(5, 6): [QuarterDuration(3, 6), QuarterDuration(2, 6)],
        QuarterDuration(5, 7): [QuarterDuration(3, 7), QuarterDuration(2, 7)],
        QuarterDuration(5, 8): [QuarterDuration(4, 8), QuarterDuration(1, 8)],
        QuarterDuration(5, 9): [QuarterDuration(3, 9), QuarterDuration(2, 9)],
        QuarterDuration(5, 11): [QuarterDuration(4, 11), QuarterDuration(1, 11)],

        QuarterDuration(7, 8): [QuarterDuration(4, 8), QuarterDuration(3, 8)],
        QuarterDuration(7, 9): [QuarterDuration(4, 9), QuarterDuration(3, 9)],
        QuarterDuration(7, 10): [QuarterDuration(5, 10), QuarterDuration(2, 10)],
        QuarterDuration(7, 11): [QuarterDuration(4, 11), QuarterDuration(3, 11)],
    },
    QuarterDuration(1, 8): {
        QuarterDuration(4, 8): [QuarterDuration(3, 8), QuarterDuration(1, 8)],
        QuarterDuration(5, 8): [QuarterDuration(3, 8), QuarterDuration(2, 8)],
        QuarterDuration(6, 8): [QuarterDuration(3, 8), QuarterDuration(3, 8)],
        QuarterDuration(7, 8): [QuarterDuration(3, 8), QuarterDuration(4, 8)],
    },
    QuarterDuration(2, 8): {
        QuarterDuration(3, 8): [QuarterDuration(2, 8), QuarterDuration(1, 8)],
        QuarterDuration(5, 8): [QuarterDuration(2, 8), QuarterDuration(3, 8)],
    },
    QuarterDuration(3, 8): {
        QuarterDuration(2, 8): [QuarterDuration(1, 8), QuarterDuration(1, 8)],
        QuarterDuration(3, 8): [QuarterDuration(1, 8), QuarterDuration(2, 8)],
        QuarterDuration(4, 8): [QuarterDuration(1, 8), QuarterDuration(3, 8)],
        QuarterDuration(5, 8): [QuarterDuration(1, 8), QuarterDuration(4, 8)],
    },
    QuarterDuration(1, 7): {
        QuarterDuration(5, 7): [QuarterDuration(3, 7), QuarterDuration(2, 7)],
        QuarterDuration(6, 7): [QuarterDuration(3, 7), QuarterDuration(3, 7)],
    },
    QuarterDuration(2, 7): {
        QuarterDuration(5, 7): [QuarterDuration(3, 7), QuarterDuration(2, 7)],
    },
    QuarterDuration(1, 6): {
        QuarterDuration(4, 6): [QuarterDuration(2, 6), QuarterDuration(2, 6)],
        QuarterDuration(5, 6): [QuarterDuration(2, 6), QuarterDuration(3, 6)],
    },
    QuarterDuration(2, 6): {
        QuarterDuration(3, 6): [QuarterDuration(2, 6), QuarterDuration(1, 6)],
        QuarterDuration(5, 6): [QuarterDuration(2, 6), QuarterDuration(3, 6)],
    },
}


def _find_nearest_quantized_value(quantized_values, values):
    output = []
    for value in values:
        nearest_quantized = min(enumerate(quantized_values), key=lambda x: abs(x[1] - value))[1]
        delta = nearest_quantized - value
        output.append((nearest_quantized, delta))
    return output


def _find_q_delta(quantized_locations, values):
    qs = _find_nearest_quantized_value(quantized_locations, values)
    d = 0
    for q in qs:
        d += abs(q[1])
    return d


def _find_quantized_locations(duration, subdivision):
    output = range(subdivision + 1)
    fr = duration / subdivision
    output = [x * fr for x in output]
    return output


def _beam_chord_group(chord_group: List['Chord']) -> None:
    """
    Function for setting beams of a list of chords (chord_group). This function is used to create or _update beams inside a beat.

    :param chord_group:
    :return: None
    """
    chord_group = [ch for ch in chord_group if ch.quarter_duration != 0]

    def add_beam(chord, number, value):
        if value == 'hook':
            if chord.quarter_duration == QuarterDuration(1, 6) and chord.offset == QuarterDuration(1, 3):
                value = 'backward hook'
            else:
                value = 'forward hook'
        for note in chord.notes:
            note.xml_object.add_child(XMLBeam(number=number, value_=value))

    def add_last_beam(chord, last_beam, current_beams, cont=False):
        if last_beam <= current_beams:
            if cont:
                add_beam(chord, 1, 'continue')
                for n in range(2, last_beam + 1):
                    add_beam(chord, n, 'end')
            else:
                for n in range(1, last_beam + 1):
                    add_beam(chord, n, 'end')
        else:
            if current_beams != 0:
                if cont:
                    add_beam(chord, 1, 'continue')
                    for n in range(2, current_beams + 1):
                        add_beam(chord, n, 'end')
                else:
                    for n in range(1, current_beams + 1):
                        add_beam(chord, n, 'end')
                for n in range(current_beams + 1, last_beam + 1):
                    add_beam(chord, n, 'backward hook')

    beams = {'eighth': 1, '16th': 2, '32nd': 3, '64th': 4, '128th': 5}
    current_beams = 0
    for index in range(len(chord_group) - 1):
        chord = chord_group[index]
        next_chord = chord_group[index + 1]
        t1, t2 = chord.notes[0].xml_type.value_, next_chord.notes[0].xml_type.value_
        b1, b2 = beams.get(t1), beams.get(t2)
        types = []
        if b1 and b2:
            if next_chord.offset == QuarterDuration(1, 2) \
                    and current_beams != 0 \
                    and (b1 == 3 or b2 == 3
                         or current_beams == 3
                         or chord.quarter_duration == QuarterDuration(3, 8)
                         or next_chord.quarter_duration == QuarterDuration(3, 8)):
                add_last_beam(chord, b1, current_beams, True)
                current_beams = 1
            elif b2 < b1 <= current_beams:
                types.append(('continue', 0, b2))
                types.append(('end', b2, current_beams))
                current_beams = b1
            elif b2 < b1 > current_beams:
                if current_beams == 0:
                    types.append(('begin', 0, b2))
                else:
                    types.append(('continue', 0, current_beams))
                    types.append(('begin', current_beams, b2))
                    types.append(('hook', b2, b1))
                current_beams = b1
            elif b2 == b1 <= current_beams:
                types.append(('continue', 0, b1))
                current_beams = b1

            elif b2 == b1 > current_beams:
                if current_beams == 0:
                    types.append(('begin', 0, b2))
                else:
                    types.append(('continue', 0, current_beams))
                    types.append(('begin', current_beams, b2))
                current_beams = b1

            elif b2 > b1 <= current_beams:
                types.append(('continue', 0, b1))
                current_beams = b1

            elif b2 > b1 > current_beams:
                if current_beams == 0:
                    types.append(('begin', 0, b1))
                else:
                    types.append(('continue', 0, current_beams))
                    types.append(('begin', current_beams, b1))
                current_beams = b1
        elif b1 and not b2:
            add_last_beam(chord, b1, current_beams)
        else:
            pass
        for l in types:
            for n in range(l[1] + 1, l[2] + 1):
                add_beam(chord, n, l[0])
        if index == len(chord_group) - 2 and b2:
            add_last_beam(next_chord, b2, current_beams)


class Beat(MusicTree, QuarterDurationMixin, QuantizeMixin, FinalizeMixin):
    """
    Beat is the direct ancestor of chords. Each :obj:`~musicscore.chord.Chord` is placed with an offset between 0 and beat's
    quarter duration inside the beat as its child .

    Quarter duration of a beat's :obj:`~musicscore.chord.Chord` child can exceed its own quarter duration. If a
    :obj:`~musicscore.chord.Chord` is longer than the quarter duration of beat's parent :obj:`~musicscore.voice.Voice`,
    a leftover :obj:`~musicscore.chord.Chord` will be added as leftover property to the :obj:`~musicscore.voice.Voice` which will be added
    to next measure's appropriate voice .

    Beat manages splitting of each child :obj:`~musicscore.chord.Chord` into appropriate tied :obj:`~musicscore.chord.Chord` s if needed,
    for example if this chord has a non-writable quarter duration like 5/6.

    The dots and tuplets are also added here to :obj:`~musicscore.chord.Chord` or directly to their :obj:`~musicscore.note.Note` children.

    Beaming and quantization are also further important tasks of a beat.
    """

    _PERMITTED_DURATIONS = {4, 2, 1, 0.5}

    def __init__(self, quarter_duration=1):
        super().__init__(quarter_duration=quarter_duration)
        self._filled_quarter_duration = 0
        self.leftover_chord = None

    def _add_child(self, child):
        child._parent = self
        self._children.append(child)
        if self.up.up.up.up:
            self.up.up.up.up.set_current_measure(staff_number=self.up.up.number, voice_number=self.up.number,
                                                 measure=self.up.up.up)

    def _add_chord(self, chord=None):
        if chord is None:
            chord = Chord(midis=60, quarter_duration=self.quarter_duration)
        return self.add_child(chord)

    def _change_children_quarter_durations(self, quarter_durations):
        if len(quarter_durations) != len(self.get_children()):
            raise ValueError
        if sum(quarter_durations) != self.quarter_duration:
            raise ValueError
        for qd, ch in zip(quarter_durations, self.get_children()):
            ch._quarter_duration = qd

    def _check_permitted_duration(self, val):
        for d in self._PERMITTED_DURATIONS:
            if val == d:
                return
        raise BeatWrongDurationError(f"Beat's quarter duration {val} is not allowed.")

    def _get_quantized_locations(self, subdivision):
        return _find_quantized_locations(self.quarter_duration, subdivision)

    def _get_quantized_quarter_durations(self, quarter_durations):
        if sum(quarter_durations) != self.quarter_duration:
            raise ValueError(
                f"Sum of quarter_durations '{quarter_durations}: {sum(quarter_durations)}' is not equal to beat quarter_duration "
                f"'{self.quarter_duration}'")

        def _get_positions():
            output = [0]
            for i, qd in enumerate(quarter_durations):
                output.append(output[i] + qd)
            return output

        positions = _get_positions()
        permitted_divs = self.get_possible_subdivisions()[:]
        best_div = permitted_divs.pop(0)
        last_q_delta = _find_q_delta(self._get_quantized_locations(subdivision=best_div), positions)

        for div in permitted_divs:
            current_q_delta = _find_q_delta(self._get_quantized_locations(subdivision=div), positions)

            if current_q_delta < last_q_delta:
                best_div = div
                last_q_delta = current_q_delta

            elif (current_q_delta == last_q_delta) and (div < best_div):
                best_div = div

        quantized_positions = [f[0] for f in
                               _find_nearest_quantized_value(self._get_quantized_locations(subdivision=best_div),
                                                             positions)]

        quantized_durations = []

        for i in range(len(quarter_durations)):
            fr = Fraction(
                quantized_positions[i + 1] - quantized_positions[i]).limit_denominator(
                trunc(best_div / self.quarter_duration))
            quantized_durations.append(QuarterDuration(fr))
        return quantized_durations

    @staticmethod
    def _split_chord(chord, quarter_durations):
        output = [chord]
        chord._quarter_duration = quarter_durations[0]
        for qd in quarter_durations[1:]:
            copied = _split_copy(chord, qd)
            output.append(copied)
        for index, ch in enumerate(output[:-1]):
            next_ch = output[index + 1]
            chord.add_tie('start')
            next_ch.add_tie('stop')
            for midi in next_ch.midis:
                midi.accidental.show = False
        return output

    def _split_not_writable(self, chord, offset):
        if SPLITTABLES.get(offset):
            quarter_durations = SPLITTABLES.get(offset).get(chord.quarter_duration)
            if quarter_durations:
                return self._split_chord(chord, quarter_durations)

    def _update_dots(self, chord_group, actual_notes):
        for note in [n for ch in chord_group for n in ch.get_children()]:
            if note.number_of_dots is None:
                if note.quarter_duration != 0:
                    if note.quarter_duration == Fraction(1, 2) and actual_notes == 6:
                        note.update_dots(number_of_dots=1)
                    else:
                        note.update_dots(note.quarter_duration.get_number_of_dots())

    def _update_tuplets(self, chord_group, actual_notes, factor=1):
        def add_bracket_to_notes(chord, type_, number=1):
            for note in chord.notes:
                if not note.xml_notations:
                    note.xml_notations = XMLNotations()
                t = note.xml_notations.xml_tuplet = XMLTuplet()
                if type_ == 'start':
                    t.bracket = 'yes'
                t.number = number
                t.type = type_

        normals = {3: 2, 5: 4, 6: 4, 7: 4, 9: 8, 10: 8, 11: 8, 12: 8, 13: 8, 14: 8, 15: 8}
        types = {8: '32nd', 4: '16th', 2: 'eighth', 1: 'quarter', 0.5: 'half'}

        actual_notes *= factor
        if int(actual_notes) != actual_notes:
            raise ValueError
        actual_notes = int(actual_notes)
        if actual_notes in normals:
            normal = normals[actual_notes]
            type_ = types[(normal / factor / self.quarter_duration.value)]
            for chord in chord_group:
                for note in chord.notes:
                    note.xml_time_modification = XMLTimeModification()
                    note.xml_time_modification.xml_actual_notes = actual_notes
                    note.xml_time_modification.xml_normal_notes = normal
                    note.xml_time_modification.xml_normal_type = type_
                if chord == chord_group[0]:
                    add_bracket_to_notes(chord, type_='start')
                elif chord == chord_group[-1]:
                    add_bracket_to_notes(chord, type_='stop')
                else:
                    pass

    def _update_note_tuplets_and_dots(self):
        actual_notes = self._get_actual_notes(self.get_children())
        if not actual_notes:
            if self.quarter_duration == 1:
                grouped_chords = _group_chords(self.get_children(), [1 / 2, 1 / 2])
                if grouped_chords:
                    for g in grouped_chords:
                        actual_notes = self._get_actual_notes(g)
                        self._update_tuplets(g, actual_notes, 1 / 2)
                        self._update_dots(g, actual_notes)
                    return
                else:
                    raise NotImplementedError(
                        'Beat cannot be halved. It cannot manage the necessary grouping of chords.')
            else:
                raise NotImplementedError(
                    'Beat with quarter_duration other than one cannot manage more than one group of chords.')

        self._update_tuplets(self.get_children(), actual_notes)
        self._update_dots(self.get_children(), actual_notes)

    def _update_note_beams(self):
        if self.get_children():
            _beam_chord_group(chord_group=self.get_children())

    @staticmethod
    def _get_actual_notes(chords):
        denominators = list(dict.fromkeys([ch.quarter_duration.denominator for ch in chords]))
        if len(denominators) > 1:
            l_c_m = lcm(denominators)
            if l_c_m not in denominators:
                return None
            else:
                return l_c_m
        else:
            return next(iter(denominators))

    def _remove_zero_quarter_durations(self):
        zeros = [ch for ch in self.get_children() if ch.quarter_duration == 0]
        for ch in zeros:
            if ch.all_midis_are_tied_to_next:
                pass
            elif ch.all_midis_are_tied_to_previous:
                ch.up.remove(ch)
            else:
                pass

    def _split_not_writable_chords(self) -> None:
        """
        This method checks if the quarter duration of all children chords must be split according to :obj:`~musicscore.beat.SPLITTABLES`
        dictionary. If chord's offset and its quarter duration exist in the dictionary a list of splitting quarter durations can be
        accessed like this: ``SPLITTABLES[chord.offset[chord.quarter_duration]]`` This dictionary can be manipulated by user during runtime
        if needed. Be careful with not writable quarter durations which have to be split (for example 5/6 must be split to 3/6,
        2/6 or some other writable quarter durations).

        :obj:`~musicscore.measure.Measure.finalize()` loops over all its beats calls this method.
        """
        for chord in self.get_children()[:]:
            split = self._split_not_writable(chord, chord.offset)
            if split:
                for ch in split:
                    ch._parent = self
                if chord == self.get_children()[-1]:
                    self._children = self.get_children()[:-1] + split
                else:
                    index = self.get_children().index(chord)
                    self._children = self.get_children()[:index] + split + self.get_children()[index + 1:]

    @property
    def is_filled(self) -> bool:
        """
        :return: ``True`` if no children can be added anymore. If ``False`` there is still room for further child or children.
        :rtype: bool
        """
        if self.filled_quarter_duration == self.quarter_duration:
            return True
        else:
            return False

    @property
    def filled_quarter_duration(self):
        """
        :return: How much of beat's quarter duration is already filled.
        :rtype: QuarterDuration
        """
        return self._filled_quarter_duration

    @property
    def number(self) -> int:
        """
        :return: Beat's number inside its parent's :obj:`musicscore.voice.Voice`
        :rtype: int
        """
        return self.up.get_children().index(self) + 1

    @property
    def offset(self) -> QuarterDuration:
        """
        :return: Offset in Beat's parent :obj:`musicscore.voice.Voice`
        :rtype: QuarterDuration
        """
        if not self.up:
            return None
        elif self.previous is None:
            return 0
        else:
            return self.previous.offset + self.previous.quarter_duration

    def add_child(self, child: Chord) -> List['Chord']:
        """
        If child's quarter duration is less than beat's remaining quarter duration: child is added to the beat.

        If child's quarter duration is greater than beat's remaining quarter duration: :obj:`~musicscore.chord.Chord`'s :obj:`~musicscore.chord.Chord.split_and_add_beatwise` is
        method called. It is possible to add a chord with a quarter duration exceeding the beat's quarter duration without splitting the chord.
        For example if the first beat in a 4/4 measure gets a chord with quarter duration 3, the chord will be added to this first beat as a
        child and the following two beats will be set to filled without having a child themselves and the parent
        :obj:`~musicscore.voice.Voice` returns the fourth beat if its :obj:`~musicscore.voice.Voice.get_current_beat` is called.

        If child's quarter duration exceeds the :obj:`~musicscore.voice.Voice`'s remaining quarter duration a leftover :obj:`~musicscore.chord.Chord` will be added to the voice and can be
        accessed when the next :obj:`~musicscore.measure.Measure` is created.

        :param child: :obj:`~musicscore.chord.Chord` to be added as child
        :return: list of split chords
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_child')
        self._check_child_to_be_added(child)
        if not self.up:
            raise BeatHasNoParentError('A child Chord can only be added to a beat if it has a voice parent.')
        if child.quarter_duration is None:
            raise ChordHasNoQuarterDurationError('Chord with no quarter_duration cannot be added to Beat.')
        if not child.midis:
            raise ChordHasNoMidisError('Chord with no midis cannot be added to Beat.')
        if self.is_filled and child.quarter_duration != 0:
            raise BeatIsFullError()
        diff = child.quarter_duration - (self.quarter_duration - self.filled_quarter_duration)
        if diff <= 0:
            self._filled_quarter_duration += child.quarter_duration
            self._add_child(child)
            return [child]
        else:
            if child.split:
                remaining_quarter_duration = child.quarter_duration
                current_beat = self
                while remaining_quarter_duration and current_beat:
                    if current_beat.quarter_duration < remaining_quarter_duration:
                        current_beat._filled_quarter_duration += current_beat.quarter_duration
                        remaining_quarter_duration -= current_beat.quarter_duration
                        current_beat = current_beat.next
                    else:
                        current_beat._filled_quarter_duration += remaining_quarter_duration
                        break
                self._add_child(child)
                return [child]
            else:
                beats = self.up.get_children()[self.up.get_children().index(self):]
                return child.split_and_add_beatwise(beats)

    def add_chord(self, *args, **kwargs):
        raise AddChordError

    def finalize(self):
        """
        finalize can only be called once.

        - It calls :obj:`~musicscore.chord.Chord.finalize()` method of all :obj:`~musicscore.chord.Chord` children.
        - Following updates are triggered: update_note_tuplets_and_dots, update_note_beams, quantize_quarter_durations (if get_quantized is
          True), _split_not_writable_chords
        """
        if self._finalized:
            raise AlreadyFinalizedError(self)
        if self.is_filled is False:
            BeatNotFullError()

        if self.get_children():
            for chord in self.get_children():
                chord.finalize()
            self._update_note_tuplets_and_dots()
            self._update_note_beams()

        self._finalized = True

    def get_children(self) -> List[Chord]:
        """
        :return: list of added children.
        :rtype: List[:obj:`~musicscore.chord.Chord`]
        """
        return super().get_children()

    def get_parent(self) -> 'Voice':
        """
        :return: parent
        :rtype: :obj:`~musicscore.voice.Voice`
        """
        return super().get_parent()

    def quantize_quarter_durations(self):
        """
        When called the positioning of children will be quantized according to :obj:`~musicscore.quantize.QuantizeMixin.get_possible_subdivisions()`
        This method is called by :obj:`~musicscore.measure.Measure`

        """
        if self.get_possible_subdivisions() and self.get_children():
            if self._get_actual_notes(self.get_children()) in self.get_possible_subdivisions():
                pass
            else:
                quarter_durations = [chord.quarter_duration for chord in self.get_children()]
                if len([d for d in quarter_durations if d != 0]) > 1:
                    self._change_children_quarter_durations(self._get_quantized_quarter_durations(quarter_durations))
                    self._remove_zero_quarter_durations()



from musicscore import Score, Chord
from musicscore.lyrics import Lyrics
from musicscore.tests.util import IdTestCase, _generate_xml_lyric
from musicscore.util import _generate_lyrics
from musicxml import XMLLyric, XMLText, XMLElision, XMLSyllabic
from musicxml.exceptions import XMLElementChildrenRequired

test_lyrics_string = ['Bla!', 'Hello World!', 'Clou-dy day.', 'Tra-la-la Ja! Tra-ra! Bah!',
                      'No-body is per-fect! - - Are they?', 'No- - - -body is per- - fect! - - Are they?']


def chord_lyric_assertions(chords, expected, number=1):
    for xl1, ch in zip(expected, chords):
        if xl1:
            xl2 = ch._xml_lyrics[number - 1]
            try:
                assert xl1.to_string() == xl2.to_string()
            except XMLElementChildrenRequired:
                xl1.xsd_check = False
                assert xl1.to_string() == xl2.to_string()
        else:
            assert not ch._xml_lyrics


class TestLyrics(IdTestCase):

    def setUp(self):
        super().setUp()
        self.score = Score()
        self.part = self.score.add_part('p1')

    def test_generate_lyrics_one_word(self):
        xml_lyrics = _generate_lyrics('Bla!')
        expected = _generate_xml_lyric(text='Bla!', syllabic='single')
        assert xml_lyrics[0].to_string() == expected.to_string()

        xml_lyrics = _generate_lyrics(['Bla!'])
        expected = _generate_xml_lyric(text='Bla!', syllabic='single')
        assert xml_lyrics[0].to_string() == expected.to_string()

    def test_generate_lyrics_list_words(self):
        words = ['Hello', 'World!']
        xml_lyrics = _generate_lyrics(words)
        expected = [_generate_xml_lyric(text=t, syllabic='single') for t in words]

        for xl1, xl2 in zip(xml_lyrics, expected):
            assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_two_syllabic_without_None(self):
        word = [('Syl', 'lable')]
        expected = [
            _generate_xml_lyric(text='Syl', syllabic='begin'),
            _generate_xml_lyric(text='lable', syllabic='end'),
        ]

        xml_lyrics = _generate_lyrics(word)
        for xl1, xl2 in zip(expected, xml_lyrics):
            assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_syllabic_without_None(self):
        word = [('Clou', 'di', 'ness')]
        expected = [
            _generate_xml_lyric(text='Clou', syllabic='begin'),
            _generate_xml_lyric(text='di', syllabic='middle'),
            _generate_xml_lyric(text='ness', syllabic='end'),
        ]

        xml_lyrics = _generate_lyrics(word)
        for xl1, xl2 in zip(expected, xml_lyrics):
            assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_list_syllabic_without_None(self):
        words = [('Tra', 'la', 'la'), 'Ja!', ('Tra', 'ra!'), 'Bah']

        expected = [
            _generate_xml_lyric(text='Tra', syllabic='begin'),
            _generate_xml_lyric(text='la', syllabic='middle'),
            _generate_xml_lyric(text='la', syllabic='end'),
            _generate_xml_lyric(text='Ja!', syllabic='single'),
            _generate_xml_lyric(text='Tra', syllabic='begin'),
            _generate_xml_lyric(text='ra!', syllabic='end'),
            _generate_xml_lyric(text='Bah', syllabic='single'),
        ]

        xml_lyrics = _generate_lyrics(words)

        for xl1, xl2 in zip(expected, xml_lyrics):
            assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_syllabic_with_None_in_between(self):
        word = ('per', None, None, 'fect!')

        expected = [
            _generate_xml_lyric(text='per', syllabic='begin'),
            None,
            None,
            _generate_xml_lyric(text='fect!', syllabic='end'),
        ]

        xml_lyrics = _generate_lyrics(word)

        for xl1, xl2 in zip(expected, xml_lyrics):
            if xl1:
                assert xl1.to_string() == xl2.to_string()
            else:
                assert xl1 == xl2 is None

    def test_generate_lyrics_single_syllabic_with_None_at_the_end(self):
        word = ('ja!', None)

        expected = [
            _generate_xml_lyric(text='ja!', syllabic='single', extend='start'),
            _generate_xml_lyric(extend='stop'),
        ]

        xml_lyrics = _generate_lyrics(word)

        for xl1, xl2 in zip(expected, xml_lyrics):
            try:
                assert xl1.to_string() == xl2.to_string()
            except XMLElementChildrenRequired:
                xl1.xsd_check = False
                assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_syllabic_with_multiple_None_at_the_end(self):
        word = ('per', 'fect!', None, None)

        expected = [
            _generate_xml_lyric(text='per', syllabic='begin'),
            _generate_xml_lyric(text='fect!', syllabic='end', extend='start'),
            _generate_xml_lyric(extend='continue'),
            _generate_xml_lyric(extend='stop'),
        ]

        xml_lyrics = _generate_lyrics(word)

        for xl1, xl2 in zip(expected, xml_lyrics):
            try:
                assert xl1.to_string() == xl2.to_string()
            except XMLElementChildrenRequired:
                xl1.xsd_check = False
                assert xl1.to_string() == xl2.to_string()

    def test_generate_lyrics_list_syllabic_with_None_in_between_and_at_the_end(self):
        words = [('No', None, None, 'body'), 'is', ('per', None, None, 'fect!', None, None), 'Are', 'they?']
        expected = [
            _generate_xml_lyric(text='No', syllabic='begin'),
            None,
            None,
            _generate_xml_lyric(text='body', syllabic='end'),
            _generate_xml_lyric(text='is', syllabic='single'),
            _generate_xml_lyric(text='per', syllabic='begin'),
            None,
            None,
            _generate_xml_lyric(text='fect!', syllabic='end', extend='start'),
            _generate_xml_lyric(extend='continue'),
            _generate_xml_lyric(extend='stop'),

            _generate_xml_lyric(text='Are', syllabic='single'),
            _generate_xml_lyric(text='they?', syllabic='single'),
        ]

        xml_lyrics = _generate_lyrics(words)

        for xl1, xl2 in zip(expected, xml_lyrics):
            if xl1:
                try:
                    assert xl1.to_string() == xl2.to_string()
                except XMLElementChildrenRequired:
                    xl1.xsd_check = False
                    assert xl1.to_string() == xl2.to_string()
            else:
                assert xl1 == xl2 is None

    def test_generate_lyrics_list_syllabic_with_None_in_between_and_at_the_end_add_to_chords(self):
        words = [('No', None, None, 'body'), 'is', ('per', None, None, 'fect!', None, None), 'Are', 'they?']
        expected = [
            _generate_xml_lyric(text='No', syllabic='begin'),
            None,
            None,
            _generate_xml_lyric(text='body', syllabic='end'),
            _generate_xml_lyric(text='is', syllabic='single'),
            _generate_xml_lyric(text='per', syllabic='begin'),
            None,
            None,
            _generate_xml_lyric(text='fect!', syllabic='end', extend='start'),
            _generate_xml_lyric(extend='continue'),
            _generate_xml_lyric(extend='stop'),

            _generate_xml_lyric(text='Are', syllabic='single'),
            _generate_xml_lyric(text='they?', syllabic='single'),
        ]

        chords = [Chord(60, 1) for _ in range(len(expected))]
        Lyrics(words).add_to_chords(chords)

        chord_lyric_assertions(chords, expected)

    def test_generate_lyrics_list_syllabic_multiple(self):
        lyrics_1 = [('syl', 'labic')]
        lyrics_2 = [('SYL', 'LABIC')]
        expected_1 = [
            _generate_xml_lyric(text='syl', syllabic='begin'),
            _generate_xml_lyric(text='labic', syllabic='end'),
        ]
        expected_2 = [
            _generate_xml_lyric(text='SYL', syllabic='begin', number=2),
            _generate_xml_lyric(text='LABIC', syllabic='end', number=2),
        ]
        chords = [Chord(60, 1) for _ in range(2)]
        Lyrics(lyrics_1).add_to_chords(chords)
        Lyrics(lyrics_2, number=2).add_to_chords(chords)
        chord_lyric_assertions(chords, expected_1)
        chord_lyric_assertions(chords, expected_2, 2)

    def test_generate_lyrics_list_syllabic_multiple_show_number(self):
        lyrics_1 = [('syl', 'labic')]
        lyrics_2 = [('SYL', 'LABIC')]
        first_syllable_1 = XMLLyric(number='1')
        first_syllable_1.add_child(XMLSyllabic('single'))
        first_syllable_1.add_child(XMLText('1.'))
        first_syllable_1.add_child(XMLElision(' '))
        first_syllable_1.add_child(XMLSyllabic('begin'))
        first_syllable_1.add_child(XMLText('syl'))

        first_syllable_2 = XMLLyric(number='2')
        first_syllable_2.add_child(XMLSyllabic('single'))
        first_syllable_2.add_child(XMLText('2.'))
        first_syllable_2.add_child(XMLElision(' '))
        first_syllable_2.add_child(XMLSyllabic('begin'))
        first_syllable_2.add_child(XMLText('SYL'))

        expected_1 = [
            first_syllable_1,
            _generate_xml_lyric(text='labic', syllabic='end'),
        ]
        expected_2 = [
            first_syllable_2,
            _generate_xml_lyric(text='LABIC', syllabic='end', number=2),
        ]
        chords = [Chord(60, 1) for _ in range(2)]
        Lyrics(lyrics_1, show_number=True).add_to_chords(chords)
        Lyrics(lyrics_2, number=2, show_number=True).add_to_chords(chords)
        chord_lyric_assertions(chords, expected_1)
        chord_lyric_assertions(chords, expected_2, 2)

    def test_generate_lyrics_list_with_None(self):
        chords = [Chord(60, 1) for _ in range(8)]
        Lyrics(['There', 'are', 'two', 'pauses!', None, None, 'And', 'go!']).add_to_chords(chords)
        assert chords[4].xml_lyrics == []
        assert chords[5].xml_lyrics == []

    def test_generate_lyrics_with_kwargs(self):
        words = [('No', None, None, 'body'), 'is', ('per', None, None, 'fect!', None, None), 'Are', 'they?']
        expected = [
            _generate_xml_lyric(text='No', syllabic='begin', default_y=-10),
            None,
            None,
            _generate_xml_lyric(text='body', syllabic='end', default_y=-10),
            _generate_xml_lyric(text='is', syllabic='single', default_y=-10),
            _generate_xml_lyric(text='per', syllabic='begin', default_y=-10),
            None,
            None,
            _generate_xml_lyric(text='fect!', syllabic='end', extend='start', default_y=-10),
            _generate_xml_lyric(extend='continue', default_y=-10),
            _generate_xml_lyric(extend='stop', default_y=-10),

            _generate_xml_lyric(text='Are', syllabic='single', default_y=-10),
            _generate_xml_lyric(text='they?', syllabic='single', default_y=-10),
        ]

        chords = [Chord(60, 1) for _ in range(len(expected))]
        Lyrics(words, default_y=-10).add_to_chords(chords)
        chord_lyric_assertions(chords, expected)



import math
import xml.etree.ElementTree as ET
from pathlib import Path

import xmltodict

from musicscore import Time, SimpleFormat, BassClef, TrebleClef
from musicscore.chord import Chord
from musicscore.exceptions import IdHasAlreadyParentOfSameTypeError, IdWithSameValueExistsError
from musicscore.key import Key
from musicscore.measure import Measure
from musicscore.part import Part, ScorePart, Id
from musicscore.quarterduration import QuarterDuration
from musicscore.score import Score
from musicscore.tests.util import IdTestCase, get_xml_elements_diff, XMLsDifferException, get_xml_diff_part, \
    generate_xml_file


class TestId(IdTestCase):

    def test_id_refs(self):
        id1 = Id('p1')
        id2 = Id('p2')
        assert Id.__refs__ == [id1, id2]
        id3 = Id('p3')
        assert Id.__refs__ == [id1, id2, id3]
        id3.delete()
        assert Id.__refs__ == [id1, id2]
        Id('p3')

    def test_id_unique(self):
        Id('p1')
        id2 = Id('p2')
        with self.assertRaises(IdWithSameValueExistsError):
            Id('p1')
        id3 = Id('p3')
        with self.assertRaises(IdWithSameValueExistsError):
            Id('p2')
        with self.assertRaises(IdWithSameValueExistsError):
            id3.value = 'p2'
        assert id3.value == 'p3'
        id2.delete()
        id3.value = 'p2'
        assert id3.value == 'p2'

    def test_id_parents(self):
        id_ = Id('p1')
        assert id_.get_parents() == []
        p = Part(id=id_)
        assert id_.get_parents() == [p, p.score_part]
        with self.assertRaises(IdHasAlreadyParentOfSameTypeError):
            Part(id=id_)
        with self.assertRaises(IdHasAlreadyParentOfSameTypeError):
            ScorePart(part=p)
        assert p.xml_object.id == p.score_part.xml_object.id == 'p1'
        id_.value = 'p2'
        assert p.xml_object.id == p.score_part.xml_object.id == 'p2'
        p.id_ = 'p3'
        assert p.xml_object.id == p.score_part.xml_object.id == 'p3'


class TestPart(IdTestCase):
    def test_part_init(self):
        p = Part(id='p1')
        p.add_child(Measure(1))
        assert p.xml_object.id == 'p1'

    def test_part_name(self):
        p = Part(id='p1')
        assert p.name == ''
        p = Part(id='p2', name='Part 1')
        assert p.name == 'Part 1'
        p.name = None
        assert p.name == ''

    def test_part_abbreviation(self):
        p = Part(id='p1')
        assert p.abbreviation is None
        p = Part(id='p2', name='Part 1', abbreviation='p 1')
        assert p.abbreviation == 'p 1'
        p = Part(id='p3', name='Part 3')
        p.abbreviation = 'p 3'
        assert p.abbreviation == 'p 3'

    def test_part_and_score_part(self):
        p = Part(id='p1')
        assert isinstance(p.score_part, ScorePart)
        assert p.score_part.xml_object.id == p.xml_object.id

    def test_part_list_multiple_parts(self):
        score = Score()
        score.add_part('p1')
        score.add_part('p2')
        score.finalize()
        assert len(score.xml_object.xml_part_list.get_children()) == 2

    def test_add_measure(self):
        p = Part('p1')
        m = p.add_measure()
        assert p.get_children()[-1] == m
        assert m.number == 1
        assert m.time.signatures == (4, 4)
        m = p.add_measure(time=(4, 3, 2, 1))
        assert p.get_children()[-1] == m
        assert m.number == 2
        assert m.time.signatures == (4, 3, 2, 1)
        m = p.add_measure()
        assert p.get_children()[-1] == m
        assert m.number == 3
        assert m.time.signatures == (4, 3, 2, 1)

    def test_add_measure_show_time(self):
        p = Part('p1')
        m = p.add_measure()
        assert m.time.show is True
        m = p.add_measure()
        assert m.time.show is False
        m = p.add_measure(time=(3, 4))
        assert m.time.show is True

    def test_add_measure_show_key(self):
        p = Part('p1')
        m = p.add_measure()
        assert m.key.show is True
        m.finalize()
        assert m.xml_object.xml_attributes.xml_key is not None
        m = p.add_measure()
        m.finalize()
        assert m.key.show is False
        m.key = Key(fifths=1)
        assert m.key.show is True
        m = p.add_measure()
        m.finalize()
        assert m.key.fifths == 1
        assert m.key.show is False
        assert m.xml_object.xml_attributes.xml_key is None

    def test_add_measure_show_clefs(self):
        p = Part('p1')
        m = p.add_measure()
        m.add_staff(1)
        m.add_staff(2)
        assert m.clefs[0].show is True
        assert m.clefs[1].show is True
        m = p.add_measure()
        assert m.clefs[0].show is False
        assert m.clefs[1].show is False
        m = p.add_measure()
        m.clefs[0].show = True
        m.finalize()
        clefs = m.xml_object.xml_attributes.find_children('XMLClef')
        assert len(clefs) == 1
        assert clefs[0].xml_sign.value_ == 'G'
        assert clefs[0].xml_line.value_ == 2

    def test_part_add_measure_check_voice(self):
        """
        Test if Part.add_measure() adds a Measure with a Staff (number=None) and Voice (number=1)
        """
        p = Part('p1')
        m1 = p.add_measure()
        assert m1.get_staff(1) is not None
        assert m1.get_voice(staff_number=1, voice_number=1) is not None
        m1.add_staff(2)
        m2 = p.add_measure()
        m2.add_voice(staff_number=2, voice_number=2)

        assert m1.get_voice(staff_number=2, voice_number=2) is None
        assert m2.get_voice(staff_number=2, voice_number=2) is not None

    def test_part_set_current_measure(self):
        p = Part('p1')
        m1 = Measure(1)
        p.set_current_measure(1, 1, m1)
        assert p.get_current_measure(1, 1) == m1
        m2 = Measure(2)
        p.set_current_measure(1, 1, m2)
        assert p.get_current_measure(1, 1) == m2
        assert p.get_current_measure(1, 2) is None

        p = Part('p2')
        m1 = Measure(1)
        p.set_current_measure(None, 1, m1)
        assert p.get_current_measure(1, 1) == m1
        m2 = Measure(2)
        p.set_current_measure(1, 1, m2)
        assert p.get_current_measure(1, 1) == m2

    def test_part_get_current_measure_simple(self):
        p = Part('p1')
        m1 = p.add_measure()
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        m2 = p.add_measure()
        m1._add_chord(Chord(midis=60, quarter_duration=2))
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        m1._add_chord(Chord(midis=60, quarter_duration=2))
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        assert m1.get_voice(staff_number=1, voice_number=1).is_filled
        m2._add_chord(Chord(midis=60, quarter_duration=2))
        assert p.get_current_measure(staff_number=1, voice_number=1) == m2
        m3 = p.add_measure()
        m4 = p.add_measure()
        m4._add_chord(Chord(midis=60, quarter_duration=2))
        assert p.get_current_measure(staff_number=1, voice_number=1) == m4

    def test_part_get_current_measure_complex(self):
        p = Part('p1')
        assert p.get_current_measure(staff_number=None, voice_number=1) is None
        m1 = p.add_measure()
        m1._add_chord(Chord(midis=60, quarter_duration=4), staff_number=2, voice_number=2)
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        assert p.get_current_measure(staff_number=1, voice_number=2) is None
        assert p.get_current_measure(staff_number=2, voice_number=1) == m1
        assert p.get_current_measure(staff_number=2, voice_number=1).get_voice(staff_number=2,
                                                                               voice_number=1).is_filled is False
        assert p.get_current_measure(staff_number=2, voice_number=2) == m1
        assert p.get_current_measure(staff_number=2, voice_number=2).get_voice(staff_number=2,
                                                                               voice_number=2).is_filled is True
        m1.add_voice(staff_number=1, voice_number=1)
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        #
        m2 = p.add_measure()
        m2._add_chord(Chord(midis=60, quarter_duration=4), staff_number=2, voice_number=2)

        assert p.get_current_measure(staff_number=1, voice_number=1) == m1
        assert p.get_current_measure(staff_number=1, voice_number=2) is None
        assert p.get_current_measure(staff_number=2, voice_number=1) == m1
        assert p.get_current_measure(staff_number=2, voice_number=2) == m2

        m1 = Measure(1)
        m1._add_chord(Chord(midis=60, quarter_duration=1))
        p = Part('p2')
        p.add_child(m1)
        assert p.get_current_measure(staff_number=1, voice_number=1) == m1


class TestScorePart(IdTestCase):

    def test_score_part_name(self):
        p = Part(id='p1')
        assert p.score_part.xml_part_name.value_ == p.name == ''
        p.name = 'Part 1'
        assert p.score_part.xml_part_name.value_ == p.name == 'Part 1'
        p.name = None
        assert p.score_part.xml_part_name.value_ == p.name == ''

    def test_score_part_abbreviation(self):
        p = Part(id='p1')
        assert p.score_part.xml_part_abbreviation == p.abbreviation is None
        p = Part(id='p2', name='Part 1', abbreviation='p 1')
        assert p.score_part.xml_part_abbreviation.value_ == p.abbreviation == 'p 1'
        p = Part(id='p3', name='Part 3')
        p.abbreviation = 'p 3'
        assert p.score_part.xml_part_abbreviation.value_ == p.abbreviation == 'p 3'

    def test_score_part_to_string(self):
        p = Part(id='p1')
        expected = """<score-part id="p1">
  <part-name />
</score-part>
"""
        assert p.score_part.to_string() == expected

    def test_add_measure_clef(self):
        p = Part('P1')
        m1 = p.add_measure(time=(3, 4))
        m2 = p.add_measure(time=(2, 4))
        p.finalize()
        assert m1.clefs[0].show is True
        assert m2.clefs[0].show is False

    def test_add_complex_quarter_durations(self):
        s = Score()
        p = s.add_child(Part('p1'))
        quarter_durations = [QuarterDuration(145, 42), QuarterDuration(1070, 737), QuarterDuration(1121, 352),
                             QuarterDuration(960, 589),
                             QuarterDuration(178, 85), QuarterDuration(134, 945), QuarterDuration(1733, 482),
                             QuarterDuration(446, 227),
                             QuarterDuration(458, 735), QuarterDuration(943, 890), QuarterDuration(650, 739),
                             QuarterDuration(939, 338),
                             QuarterDuration(327, 227), QuarterDuration(461, 183), QuarterDuration(1, 2),
                             QuarterDuration(1123, 457),
                             QuarterDuration(24, 59), QuarterDuration(1075, 397), QuarterDuration(604, 219),
                             QuarterDuration(18, 605),
                             QuarterDuration(2398, 999), QuarterDuration(235, 608), QuarterDuration(677, 799),
                             QuarterDuration(136, 637),
                             QuarterDuration(2717, 783), QuarterDuration(5, 643)]
        # first_qd = math.ceil(sum(quarter_durations[:4])) - sum(quarter_durations[:4])
        # qds = [first_qd] + [quarter_durations[4]]
        # last_qd = math.ceil(sum(qds)) - sum(qds)
        # qds += [last_qd]
        # qds = quarter_durations[:5]
        # last_qd = math.ceil(sum(qds)) - sum(qds)
        # qds += [last_qd]
        # qds = [QuarterDuration(145, 42), QuarterDuration(1070, 737), QuarterDuration(1121, 352), QuarterDuration(960, 589),
        #        QuarterDuration(178, 85), QuarterDuration(32, 171)]
        # qds = qds[4:]
        # first_qd = math.ceil(sum(qds)) - sum(qds)
        # print(first_qd)
        # qds = [first_qd] + qds
        # print(qds)
        qds = [QuarterDuration(23, 32), QuarterDuration(178, 85), QuarterDuration(32, 171), QuarterDuration(1)]
        for qd in qds:
            p.add_chord(Chord(midis=60, quarter_duration=qd))
        expected = [QuarterDuration(23, 32), QuarterDuration(9, 32), QuarterDuration(1, 1),
                    QuarterDuration(139, 171), QuarterDuration(32, 171), QuarterDuration(1, 1)]

        assert [ch.quarter_duration for ch in p.get_chords()] == expected

        for beat in [b for m in p.get_children() for st in m.get_children() for v in st.get_children() for b in
                     v.get_children()]:
            beat.quantize_quarter_durations()
            beat._split_not_writable_chords()
        expected = [QuarterDuration(3, 7), QuarterDuration(2, 7), QuarterDuration(2, 7), QuarterDuration(1, 1),
                    QuarterDuration(4, 5), QuarterDuration(1, 5), QuarterDuration(1, 1)]
        assert [ch.quarter_duration for ch in p.get_chords()] == expected

    def test_chord_previous(self):
        p = Part('P1')
        for qd in [1, 2.5, 0.5]:
            p.add_chord(Chord(midis=60, quarter_duration=qd))

        for i in range(1, len(p.get_children())):
            current = p.get_children()[i]
            previous = p.get_children()[i - 1]
            assert current.previous == previous

    def test_chord_offsets(self):
        p = Part('P1')
        for qd in [1, 2.5, 0.5]:
            p.add_chord(Chord(midis=60, quarter_duration=qd))

        assert [ch.quarter_duration for ch in p.get_chords()] == [1, 2, 0.5, 0.5]
        assert [ch.offset for ch in p.get_chords()] == [0, 0, 0, 0.5]


class TestAddChordToPart(IdTestCase):
    def setUp(self):
        self.score = Score()
        super().setUp()

    @staticmethod
    def _get_clefs_of_measure(part, measure_number=1):
        return xmltodict.parse(part.get_children()[measure_number - 1].to_string())['measure']['attributes'].get('clef',
                                                                                                                 None)

    def test_add_one_chord_quarter_duration_4(self):
        xml_file = Path(__file__).stem + '_add_quarter_duration_4.xml'
        p = self.score.add_part(id='part1')
        chord = Chord(60, 4)
        p.add_chord(chord)
        self.score.export_xml(xml_file)
        output_score = ET.parse(xml_file)
        output_part_xml = output_score.find('part')
        output_part_xml_dict = xmltodict.parse(ET.tostring(output_part_xml))
        expected = {'part': {'@id': 'part1', 'measure': {'@number': '1',
                                                         'attributes': {'divisions': '1', 'key': {'fifths': '0'},
                                                                        'time': {'beats': '4', 'beat-type': '4'},
                                                                        'clef': {'sign': 'G', 'line': '2'}},
                                                         'note': {'pitch': {'step': 'C', 'octave': '4'},
                                                                  'duration': '4', 'voice': '1', 'type': 'whole'},
                                                         'barline': {'@location': 'right',
                                                                     'bar-style': 'light-heavy'}}}}
        assert output_part_xml_dict == expected

    def test_add_chord_to_staff_with_bass_clef(self):
        part = Part(id='part-1')
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)
        assert self._get_clefs_of_measure(part, 1)[1] == {'@number': '2', 'sign': 'F', 'line': '4'}
        assert self._get_clefs_of_measure(part, 2) is None

    def test_add_chord_to_both_staves(self):
        part = Part(id='part-1')
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=1)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=1)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)
        assert self._get_clefs_of_measure(part, 1)[1] == {'@number': '2', 'sign': 'F', 'line': '4'}
        assert self._get_clefs_of_measure(part, 2) is None

    def test_add_chord_to_both_staves_other_order(self):
        part = Part(id='part-1')
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=1)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=1)

        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)
        part.add_chord(Chord(midis=60, quarter_duration=4), staff_number=2)

        assert self._get_clefs_of_measure(part, 1)[1] == {'@number': '2', 'sign': 'F', 'line': '4'}
        assert self._get_clefs_of_measure(part, 2) is None

    def test_add_long_chord_to_staff_with_bass_clef(self):
        part = Part(id='part-1')
        part.add_chord(Chord(midis=60, quarter_duration=11), staff_number=2)
        assert self._get_clefs_of_measure(part, 1)[1] == {'@number': '2', 'sign': 'F', 'line': '4'}
        assert self._get_clefs_of_measure(part, 2) is None
        assert self._get_clefs_of_measure(part, 3) is None

    def test_different_long_chords_to_staff_with_bass_clef(self):
        part = Part(id='part-1')
        midis = list(range(60, 65))
        quarter_durations = list(range(1, 6))
        for m, q in zip(midis, quarter_durations):
            part.add_chord(Chord(m, q), staff_number=2)
        assert self._get_clefs_of_measure(part, 1)[1] == {'@number': '2', 'sign': 'F', 'line': '4'}
        for measure_number in range(2, math.ceil(sum(quarter_durations) / 4)):
            assert self._get_clefs_of_measure(part, measure_number) is None

    def test_add_chord_to_two_staves_same_length(self):
        part = Part(id='part-1')
        midis = list(range(60, 65))
        quarter_durations = list(range(1, 6))
        for m, q in zip(midis, quarter_durations):
            part.add_chord(Chord(m, q), staff_number=1)
        for m, q in zip(reversed(midis), reversed(quarter_durations)):
            part.add_chord(Chord(m, q), staff_number=2)
        assert self._get_clefs_of_measure(part, 1)[1] == {'@number': '2', 'sign': 'F', 'line': '4'}
        for measure_number in range(2, math.ceil(sum(quarter_durations) / 4)):
            assert self._get_clefs_of_measure(part, measure_number) is None

    def test_add_chord_different_staves_and_voices(self):
        p = Part('p1')
        p.add_chord(Chord(60, 1))
        m = p.get_children()[0]
        assert m.get_voice(staff_number=1, voice_number=1) is not None
        p.add_chord(Chord(61, 2), staff_number=2, voice_number=4)
        for i in range(1, 5):
            assert m.get_voice(staff_number=2, voice_number=i) is not None
        p.add_chord(Chord(62, 3), staff_number=1, voice_number=1)
        assert [ch.quarter_duration for ch in m.get_voice(staff_number=1, voice_number=1).get_chords()] == [1, 3]
        assert [ch.midis[0].value for ch in m.get_voice(staff_number=1, voice_number=1).get_chords()] == [60, 62]
        assert [ch.quarter_duration for ch in m.get_voice(staff_number=2, voice_number=4).get_chords()] == [2]

    def test_add_chord_with_leftover(self):
        p = Part('p1')
        ch = Chord(60, 5)
        p.add_chord(ch)
        assert len(p.get_children()) == 2
        m1, m2 = p.get_children()
        assert p.get_current_measure(1, 1) == m2
        assert m1.get_voice(staff_number=1, voice_number=1).is_filled
        assert not m2.get_voice(staff_number=1, voice_number=1).is_filled

    def test_add_chord_to_full_measure(self):
        p = Part('p1')
        p.add_chord(Chord(60, 4))
        p.add_chord(Chord(60, 6))
        m1, m2, m3 = p.get_children()
        assert p.get_current_measure() == m3

    def test_add_chord_to_existing_measures(self):
        p = Part('p1')
        for _ in range(3):
            p.add_measure()
        m1, m2, m3 = p.get_children()
        assert p.get_current_measure() == m1
        p.add_chord(Chord(60, 5))
        assert p.get_current_measure() == m2

    def test_add_chord_with_staff_number(self):
        p = Part('P1')
        ch1 = Chord(60, 1)
        ch2 = Chord(48, 1)
        p.add_chord(ch1, staff_number=1)
        assert ch1.get_staff_number() is None
        p.add_chord(ch2, staff_number=2)
        assert ch1.get_staff_number() == 1
        assert ch2.get_staff_number() == 2
        p.finalize()
        assert ch1.notes[0].xml_staff.value_ == 1
        assert ch2.notes[0].xml_staff.value_ == 2

    def test_add_chord_with_clef(self):
        path = Path(__file__).stem + '_add_chord_with_clef.xml'
        expected = Path(__file__).stem + '_add_chord_with_clef_expected.xml'
        sf1 = SimpleFormat(midis=[60, (61, 66), 62], quarter_durations=[1, 2, 3])
        sf2 = SimpleFormat(midis=[60, (61, 66), 62], quarter_durations=[1, 2, 3])
        sf1.chords[1].clef = BassClef()

        sf2.chords[1].clef = BassClef()
        sf2.chords[2].clef = TrebleClef()
        score = Score()
        part = score.add_part(id='part-1')
        for index, simpleformat in enumerate([sf1, sf2]):
            for chord in simpleformat.chords:
                part.add_chord(chord, staff_number=index + 1)
        part.get_staff(1, 2).clef = TrebleClef()
        score.export_xml(path)

        get_xml_diff_part(expected, path, Path(__file__))

    def test_add_first_chord_with_clef(self):
        path = Path(__file__).stem + '_add_first_chord_with_clef.xml'
        expected = Path(__file__).stem + '_add_first_chord_with_clef_expected.xml'
        sf1 = SimpleFormat(midis=[60, (61, 66), 62], quarter_durations=[2, 6, 4])
        sf1.chords[0].clef = BassClef()
        sf1.chords[2].clef = TrebleClef()
        generate_xml_file(Score(), sf1, path=path)
        get_xml_diff_part(expected, path, Path(__file__))


class TestSplitQdAndTime(IdTestCase):
    def test_part_add_split_original_ties_1(self):
        chord = Chord(midis=60, quarter_duration=10)
        p = Part('p1')
        p.add_measure((5, 4))
        p.add_chord(chord)
        for chord in p.get_chords():
            assert chord._original_starting_ties == [set()]

    def test_part_add_split_original_ties_2(self):
        chord = Chord(midis=[60, 62], quarter_duration=10)
        chord.midis[0].add_tie('start')
        chord.midis[1].add_tie('stop')
        p = Part('p1')
        p.add_measure((5, 4))
        p.add_chord(chord)
        for chord in p.get_chords():
            assert chord._original_starting_ties == [{'start'}, {'stop'}]

    def test_part_add_split_chord_5_3_4(self):
        chord = Chord(midis=60, quarter_duration=5)
        p = Part('p1')
        p.add_measure((3, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        assert len(m.get_chords()) == 1
        ch = m.get_chords()[0]
        assert ch.quarter_duration == 2
        assert ch.all_midis_are_tied_to_previous
        assert not ch.all_midis_are_tied_to_next

    def test_part_add_split_chord_5_4_4(self):
        chord = Chord(midis=60, quarter_duration=5)
        p = Part('p1')
        p.add_measure((4, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        assert len(m.get_chords()) == 1
        ch = m.get_chords()[0]
        assert ch.quarter_duration == 1
        assert ch.all_midis_are_tied_to_previous
        assert not ch.all_midis_are_tied_to_next

    def test_part_add_split_chord_5_5_4(self):
        chord = Chord(midis=60, quarter_duration=5)
        p = Part('p1')
        p.add_measure((5, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 1
        m = p.get_children()[-1]
        all_chords = [ch1, ch2] = m.get_chords()
        assert [ch.quarter_duration for ch in all_chords] == [3, 2]
        assert not ch1.all_midis_are_tied_to_previous
        assert ch1.all_midis_are_tied_to_next
        assert ch2.all_midis_are_tied_to_previous
        assert not ch2.all_midis_are_tied_to_next

    def test_part_add_split_chord_6_4_4(self):
        chord = Chord(midis=60, quarter_duration=6)
        p = Part('p1')
        p.add_measure((4, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        assert len(m.get_chords()) == 1
        ch = m.get_chords()[0]
        assert ch.quarter_duration == 2
        assert ch.all_midis_are_tied_to_previous
        assert not ch.all_midis_are_tied_to_next

    def test_part_add_split_chord_6_5_4(self):
        chord = Chord(midis=60, quarter_duration=6)
        p = Part('p1')
        p.add_measure((5, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        assert len(m.get_chords()) == 1
        ch = m.get_chords()[0]
        assert ch.quarter_duration == 1
        assert ch.all_midis_are_tied_to_previous
        assert not ch.all_midis_are_tied_to_next

    def test_part_add_split_chord_7_7_4(self):
        chord = Chord(midis=60, quarter_duration=7)
        p = Part('p1')
        p.add_measure((7, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 1
        m = p.get_children()[-1]
        all_chords = [ch1, ch2] = m.get_chords()
        assert [ch.quarter_duration for ch in all_chords] == [4, 3]
        assert not ch1.all_midis_are_tied_to_previous
        assert ch1.all_midis_are_tied_to_next
        assert ch2.all_midis_are_tied_to_previous
        assert not ch2.all_midis_are_tied_to_next

    def test_part_add_split_chord_10_7_4(self):
        chord = Chord(midis=60, quarter_duration=10)
        p = Part('p1')
        p.add_measure((7, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        assert len(m.get_chords()) == 1
        ch = m.get_chords()[0]
        assert ch.quarter_duration == 3
        assert ch.all_midis_are_tied_to_previous
        assert not ch.all_midis_are_tied_to_next

    def test_part_add_split_chord_12_7_4(self):
        chord = Chord(midis=60, quarter_duration=12)
        p = Part('p1')
        p.add_measure((7, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        all_chords = [ch1, ch2] = m.get_chords()
        assert [ch.quarter_duration for ch in all_chords] == [3, 2]
        assert ch1.all_midis_are_tied_to_previous
        assert ch1.all_midis_are_tied_to_next
        assert ch2.all_midis_are_tied_to_previous
        assert not ch2.all_midis_are_tied_to_next

    def test_part_add_split_chord_10_5_4(self):
        chord = Chord(midis=60, quarter_duration=10)
        p = Part('p1')
        p.add_measure((5, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 2
        m = p.get_children()[-1]
        all_chords = [ch1, ch2] = m.get_chords()
        assert [ch.quarter_duration for ch in all_chords] == [3, 2]
        assert ch1.all_midis_are_tied_to_next
        assert ch1.all_midis_are_tied_to_previous
        assert not ch2.all_midis_are_tied_to_next
        assert ch2.all_midis_are_tied_to_previous

    def test_part_add_split_chord_10_10_4(self):
        chord = Chord(midis=60, quarter_duration=10)
        p = Part('p2')
        p.add_measure((10, 4))
        p.add_chord(chord)
        assert len(p.get_children()) == 1
        m = p.get_children()[-1]
        all_chords = [ch1, ch2, ch3] = m.get_chords()
        assert [ch.quarter_duration for ch in all_chords] == [4, 4, 2]
        assert not ch1.all_midis_are_tied_to_previous
        assert ch1.all_midis_are_tied_to_next
        assert ch2.all_midis_are_tied_to_previous
        assert ch2.all_midis_are_tied_to_next
        assert ch3.all_midis_are_tied_to_previous
        assert not ch3.all_midis_are_tied_to_next

    def test_add_chords_with_partially_tied_notes(self):
        midis = [[60, 63], [61, 63], [62, 64], [62, 65]]
        quarter_durations = [1, 2, 2, 1]
        chords = [Chord(ms, qd) for ms, qd in zip(midis, quarter_durations)]
        chords[0].midis[1].add_tie('start')
        chords[1].midis[1].add_tie('stop')
        chords[2].midis[0].add_tie('start')
        chords[3].midis[0].add_tie('stop')
        p = Part(id='part-1')
        for chord in chords:
            p.add_chord(chord)
        path = Path(__file__).stem + '_add_chords_with_partially_tied_notes.xml'
        expected = Path(__file__).stem + '_add_chords_with_partially_tied_notes_expected.xml'
        score = Score()
        score.add_child(p)
        score.export_xml(path)
        get_xml_diff_part(expected, path, Path(__file__))

    def test_add_chords_with_partially_tied_notes_simplified(self):
        midis = [[62, 64], [62, 65]]
        quarter_durations = [3, 1]
        chords = [Chord(ms, qd) for ms, qd in zip(midis, quarter_durations)]
        chords[0].midis[0].add_tie('start')
        chords[1].midis[0].add_tie('stop')
        p = Part('part-1')
        p.add_measure(Time(2, 4))
        for chord in chords:
            p.add_chord(chord)
        path = Path(__file__).stem + '_add_chords_with_partially_tied_notes_simplified.xml'
        expected = Path(__file__).stem + '_add_chords_with_partially_tied_notes_simplified_expected.xml'
        score = Score()
        score.add_child(p)
        score.export_xml(path)
        el1 = ET.parse(Path(__file__).parent / path).getroot().find("part[@id='part-1']")
        el2 = ET.parse(Path(__file__).parent / expected).getroot().find("part[@id='part-1']")
        diff = get_xml_elements_diff(el1=el1, el2=el2)
        if diff:
            raise XMLsDifferException(diff)



from pathlib import Path

from quicktions import Fraction

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import IdTestCase, generate_all_16ths, generate_all_32nds


class TestBeams1(IdTestCase):
    def test_beams_1(self):
        """
        Write all possible combinations of 16ths and 32nds in a beat
        """
        s = Score()
        p = s.add_child(Part('P1', name='Music'))

        p.add_measure(time=(1, 4))

        groups = [(Fraction(1, 2), Fraction(1, 2))] + generate_all_16ths() + generate_all_32nds()
        for group in groups:
            for qd in group:
                p.add_chord(Chord(60, qd))
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



from pathlib import Path
from unittest import TestCase

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import IdTestCase


class TestHelloWorld(IdTestCase):
    def test_export_hello_world(self):
        """
        Hello World, Variation I: C4 pitch, 20 beats long, in three 5/4 measures with treble clef.
        """
        """
        Tester creates a timewise score
        """
        s = Score()
        """
        He adds a measure with one part to it (default Measure has a 4/4 time signature)
        """
        p = s.add_child(Part('P1', name='Music'))
        """
        He adds a 5/4 measure
        """
        p.add_measure(time=(5, 4))
        """
        He adds a long Chord with midi 61 to Part
        """
        p.add_chord(Chord(60, 15))
        """
        ... and exports the xml
        """
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



from pathlib import Path

from musicscore.chord import Chord
from musicscore.clef import TrebleClef
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import IdTestCase


class TestHelloRests(IdTestCase):
    def test_export_hello_world(self):
        """
        Hallo World, Variation III: Pitches and rests: 1, 2, ..., 15 beats (quarters) first in 4/4 measures, than in 5/4 and 6/4 measures.
        Each note / rest shows its duration as lyrics.
        """
        """
        Tester creates a timewise score
        """
        s = Score()
        """
        He adds a part
        """
        p = s.add_child(Part('P1', name='Music'))
        """
        He adds a long list of chords with ascending arithmatic progression a1=1 an=15 d=1:
          measures in 4/4 time
          measures in 5/4 time
          measures in 6/4 time
          as pitches and as rests
        """
        for time in [(4, 4), (5, 4), (6, 4)]:
            m = p.add_measure(time=time)
            if time == (4, 4):
                st = m.add_staff()
                st.clef = TrebleClef()
            for qd in list(range(1, 16)):
                ch = Chord(midis=60, quarter_duration=qd)
                ch.add_lyric(qd)
                p.add_chord(chord=ch, staff_number=1)
                ch = Chord(midis=0, quarter_duration=qd)
                ch.add_lyric(qd)
                p.add_chord(chord=ch, staff_number=2)

        """
        ... and exports the xml
        """
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



from pathlib import Path
from unittest import TestCase

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import IdTestCase


class TestHelloWorld(IdTestCase):
    def test_export_hello_world(self):
        """
        Hello World Variation II: C4 with 1/3, 1/5, 1/6 and 1/10 based durations in one measure.
        """
        """
        Tester creates a timewise score
        """
        s = Score()
        """
        He adds a measure with one part to it (default Measure has a 4/4 time signature)
        """
        p = s.add_child(Part('P1', name='Music'))
        """
        He adds a 4/4 measure
        """
        p.add_measure((4, 4))
        """
        He adds a fractional Chord with midi 60 to Part
        """
        p.add_chord(Chord(60, 1 / 3))
        p.add_chord(Chord(60, 2 / 3))
        p.add_chord(Chord(60, 1 / 5))
        p.add_chord(Chord(60, 4 / 5))
        p.add_chord(Chord(60, 2 / 5))
        p.add_chord(Chord(60, 3 / 5))
        p.add_chord(Chord(60, 1 / 6))
        p.add_chord(Chord(60, 1 / 6))
        p.add_chord(Chord(60, 1 / 6))
        p.add_chord(Chord(60, 1 / 10))
        p.add_chord(Chord(60, 3 / 10))
        p.add_chord(Chord(60, 1 / 10))
        """
        ... and exports the xml
        """
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



from pathlib import Path
from unittest import TestCase, skip

from musicscore import Chord
from musicscore.exceptions import WrongNumberOfChordsError
from musicscore.midi import C

from musicscore.score import Score
from musicscore.tests.util import diff_xml, _create_expected_path, create_test_objects
from musicscore.util import lcm, isinstance_as_string, XML_DYNAMIC_CLASSES, XML_ARTICULATION_CLASSES, \
    XML_ORNAMENT_CLASSES, XML_ORNAMENT_AND_OTHER_NOTATIONS, XML_TECHNICAL_CLASSES, XML_OTHER_NOTATIONS, \
    XML_DIRECTION_TYPE_CLASSES, XML_OTHER_NOTATIONS, XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS, slur_chords, wedge_chords, \
    trill_chords, bracket_chords, octave_chords
from musicxml import XMLTrillMark, XMLWavyLine, XMLOctaveShift


class TestUtils(TestCase):
    def test_isinstance_as_string(self):
        assert isinstance_as_string(Score(), 'Score')
        assert isinstance_as_string(Score(), 'MusicTree')
        assert not isinstance_as_string(Score(), 'str')

        class Measure:
            pass

        assert not isinstance_as_string(Measure(), 'MusicTree')

        assert isinstance_as_string(C(4), 'MidiNote')
        assert isinstance_as_string(C(4), 'Midi')
        assert isinstance_as_string(C(4), 'C')
        assert not isinstance_as_string(C(4), 'str')

    def test_create_expected_path(self):
        path = Path(__file__).parent / 'test_util_diff_xml.xml'
        expected_path = _create_expected_path(path)
        assert expected_path == Path(__file__).parent / 'test_util_diff_xml_expected.xml'

    def test_diff_xml_no_diff(self):
        path = Path(__file__).parent / 'test_util_diff_xml.xml'
        assert diff_xml(path) == []

    def test_diff_xml_with_diff(self):
        path = Path(__file__).parent / 'test_util_diff_xml.xml'
        path2 = Path(__file__).parent / 'test_util_diff_xml_with_diff.xml'
        assert diff_xml(path, path2) == ['- <key>', '- <fifths>0</fifths>', '- </key>']

    def test_lcm(self):
        assert lcm([3, 4, 5, 7]) == 420
        assert lcm([2, 4, 6]) == 12

    def test_slur_chords(self):
        chords = [Chord(60, 1)]
        with self.assertRaises(WrongNumberOfChordsError):
            slur_chords(chords)
        chords.extend([Chord(61, 1), Chord(62, 1)])
        slur_chords(chords)
        assert chords[0].get_slurs()[0].type == 'start'
        assert chords[1].get_slurs()[0].type == 'continue'
        assert chords[2].get_slurs()[0].type == 'stop'

    def test_wedge_chords(self):
        chords = [Chord(60, 1)]
        with self.assertRaises(WrongNumberOfChordsError):
            wedge_chords(chords, 'crescendo')
        chords.extend([Chord(61, 1), Chord(62, 1)])
        wedge_chords(chords, 'crescendo')
        assert chords[0].get_wedges()[0].type == 'crescendo'
        assert chords[1].get_wedges()[0].type == 'continue'
        assert chords[2].get_wedges()[0].type == 'stop'

    def test_trill_chords(self):
        chords = [Chord(60, 1)]
        with self.assertRaises(WrongNumberOfChordsError):
            trill_chords(chords)
        chords.extend([Chord(61, 1), Chord(62, 1)])
        trill_chords(chords)
        assert len(chords[0].get_x(XMLTrillMark)) == 1
        assert len(chords[1].get_x(XMLTrillMark)) == 0
        assert len(chords[2].get_x(XMLTrillMark)) == 0
        assert chords[0].get_x(XMLWavyLine)[0].type == 'start'
        assert chords[1].get_x(XMLWavyLine)[0].type == 'continue'
        assert chords[2].get_x(XMLWavyLine)[0].type == 'stop'

    def test_bracket_chords(self):
        chords = [Chord(60, 1)]
        with self.assertRaises(WrongNumberOfChordsError):
            bracket_chords(chords)
        chords.extend([Chord(61, 1), Chord(62, 1)])
        bracket_chords(chords)
        assert chords[0].get_brackets()[0].type == 'start'
        assert chords[1].get_brackets()[0].type == 'continue'
        assert chords[2].get_brackets()[0].type == 'stop'

        for i, ch in enumerate(chords):
            bracket = ch.get_brackets()[0]
            assert bracket.line_type == 'solid'
            # assert bracket.placement == 'above'
            assert bracket.number == 1
            if i == 0:
                assert bracket.line_end == 'down'
            elif i == 1:
                assert bracket.line_end == 'none'
            else:
                assert bracket.line_end == 'down'

        chords = [Chord(60, 1) for _ in range(3)]
        bracket_chords(chords, 'dashed', 'none', 'up', placement='below', number=2)
        for i, ch in enumerate(chords):
            bracket = ch.get_brackets()[0]
            # assert bracket.placement == 'below'
            assert bracket.number == 2
            assert bracket.line_type == 'dashed'
            if i == 0:
                assert bracket.line_end == 'none'
            elif i == 1:
                assert bracket.line_end == 'none'
            else:
                assert bracket.line_end == 'up'

    def test_octave_chords(self):
        chords = [Chord(60, 1) for _ in range(3)]
        octave_chords(chords)
        assert chords[0].get_x(XMLOctaveShift)[0].type == 'down'
        assert chords[1].get_x(XMLOctaveShift)[0].type == 'continue'
        assert chords[2].get_x(XMLOctaveShift)[0].type == 'stop'

        chords = [Chord(60, 1) for _ in range(3)]
        octave_chords(chords, type='down', size=15, number=2)
        assert chords[0].get_x(XMLOctaveShift)[0].type == 'down'
        assert chords[1].get_x(XMLOctaveShift)[0].type == 'continue'
        assert chords[2].get_x(XMLOctaveShift)[0].type == 'stop'
        for ch in chords:
            assert ch.get_x(XMLOctaveShift)[0].size == 15
            assert ch.get_x(XMLOctaveShift)[0].number == 2


class TestTestObjects(TestCase):

    def test_direction_type_test_objects(self):
        test_object = create_test_objects(type_='direction_type')
        assert len(XML_DIRECTION_TYPE_CLASSES + XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS) == len(test_object)
        for obj in test_object:
            assert obj.__class__ in XML_DIRECTION_TYPE_CLASSES or obj.__class__ in XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS

    def test_ornament_test_objects(self):
        test_object = create_test_objects(type_='ornament')
        assert len(XML_ORNAMENT_CLASSES + XML_ORNAMENT_AND_OTHER_NOTATIONS) == len(test_object)
        for obj in test_object:
            assert obj.__class__ in XML_ORNAMENT_CLASSES + XML_ORNAMENT_AND_OTHER_NOTATIONS

    def test_technical_test_objects(self):
        test_object = create_test_objects(type_='technical')
        assert len(XML_TECHNICAL_CLASSES) == len(test_object)
        for obj in test_object:
            assert obj.__class__ in XML_TECHNICAL_CLASSES

    def test_articulation_test_objects(self):
        test_object = create_test_objects(type_='articulation')
        assert len(XML_ARTICULATION_CLASSES) == len(test_object)
        for obj in test_object:
            assert obj.__class__ in XML_ARTICULATION_CLASSES

    def test_other_notation_test_objects(self):
        test_object = create_test_objects(type_='notation')
        assert len(XML_OTHER_NOTATIONS + XML_ORNAMENT_AND_OTHER_NOTATIONS) == len(test_object)
        for obj in test_object:
            assert obj.__class__ in XML_OTHER_NOTATIONS + XML_ORNAMENT_AND_OTHER_NOTATIONS

    def test_dynamics_test_objects(self):
        test_object = create_test_objects(type_='dynamics')
        assert len(XML_DYNAMIC_CLASSES) == len(test_object)
        for obj in test_object:
            assert obj.__class__ in XML_DYNAMIC_CLASSES



"""
Some grace notes and after-graces
"""
from pathlib import Path

from musicscore import Score, Chord, A, E, G
from musicscore.tests.util import IdTestCase


class TestLily24d(IdTestCase):
    def test_lily_24d_AfterGrace(self):
        score = Score()
        part = score.add_part('p1')

        chords = [Chord(E(5), 2), Chord(E(5), 2)]
        for midi in [G(5), A(5), A(5)]:
            chords[0].add_grace_chord(midi, type_='16th', position='after')

        chords[1].add_grace_chord(G(5), type_='16th', position='after')
        chords[1].add_grace_chord(A(5), type_='16th', position='after')

        for ch in chords:
            part.add_chord(ch)

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from typing import Optional

from musicxml.xmlelement.xmlelement import XMLAccidental

from musicscore.musictree import MusicTree
from musicscore.xmlwrapper import XMLWrapper

__all__ = ['STANDARD', 'FLAT', 'SHARP', 'ENHARMONIC', 'FORCESHARP', 'FORCEFLAT', 'SIGNS', 'Accidental']
#:
STANDARD = {
    0: ('C', 0, 0),
    0.5: ('C', 0.5, 0),
    1: ('C', 1, 0),
    1.5: ('C', 1.5, 0),
    2: ('D', 0, 0),
    2.5: ('E', -1.5, 0),
    3: ('E', -1, 0),
    3.5: ('E', -0.5, 0),
    4: ('E', 0, 0),
    4.5: ('E', 0.5, 0),
    5: ('F', 0, 0),
    5.5: ('F', 0.5, 0),
    6: ('F', 1, 0),
    6.5: ('F', 1.5, 0),
    7: ('G', 0, 0),
    7.5: ('A', -1.5, 0),
    8: ('A', -1, 0),
    8.5: ('A', -0.5, 0),
    9: ('A', 0, 0),
    9.5: ('B', -1.5, 0),
    10: ('B', -1, 0),
    10.5: ('B', -0.5, 0),
    11: ('B', 0, 0),
    11.5: ('C', -0.5, 1)
}

#:
FLAT = {
    0: ('C', 0, 0),
    0.5: ('D', -1.5, 0),
    1: ('D', -1, 0),
    1.5: ('D', -0.5, 0),
    2: ('D', 0, 0),
    2.5: ('E', -1.5, 0),
    3: ('E', -1, 0),
    3.5: ('E', -0.5, 0),
    4: ('E', 0, 0),
    4.5: ('F', -0.5, 0),
    5: ('F', 0, 0),
    5.5: ('G', -1.5, 0),
    6: ('G', -1, 0),
    6.5: ('G', -0.5, 0),
    7: ('G', 0, 0),
    7.5: ('A', -1.5, 0),
    8: ('A', -1, 0),
    8.5: ('A', -0.5, 0),
    9: ('A', 0, 0),
    9.5: ('B', -1.5, 0),
    10: ('B', -1, 0),
    10.5: ('B', -0.5, 0),
    11: ('B', 0, 0),
    11.5: ('C', -0.5, 1)
}

#:
SHARP = {
    0: ('C', 0, 0),
    0.5: ('C', 0.5, 0),
    1: ('C', 1, 0),
    1.5: ('C', 1.5, 0),
    2: ('D', 0, 0),
    2.5: ('D', 0.5, 0),
    3: ('D', 1, 0),
    3.5: ('D', 1.5, 0),
    4: ('E', 0, 0),
    4.5: ('E', 0.5, 0),
    5: ('F', 0, 0),
    5.5: ('F', 0.5, 0),
    6: ('F', 1, 0),
    6.5: ('F', 1.5, 0),
    7: ('G', 0, 0),
    7.5: ('G', 0.5, 0),
    8: ('G', 1, 0),
    8.5: ('G', 1.5, 0),
    9: ('A', 0, 0),
    9.5: ('A', 0.5, 0),
    10: ('A', 1, 0),
    10.5: ('A', 1.5, 0),
    11: ('B', 0, 0),
    11.5: ('B', 0.5, 0)
}

#:
ENHARMONIC = {
    0: ('C', 0, 0),
    0.5: ('D', -1.5, 0),
    1: ('D', -1, 0),
    1.5: ('D', -0.5, 0),
    2: ('D', 0, 0),
    2.5: ('D', 0.5, 0),
    3: ('D', 1, 0),
    3.5: ('D', 1.5, 0),
    4: ('E', 0, 0),
    4.5: ('F', -0.5, 0),
    5: ('F', 0, 0),
    5.5: ('G', -1.5, 0),
    6: ('G', -1, 0),
    6.5: ('G', -0.5, 0),
    7: ('G', 0, 0),
    7.5: ('G', 0.5, 0),
    8: ('G', 1, 0),
    8.5: ('G', 1.5, 0),
    9: ('A', 0, 0),
    9.5: ('A', 0.5, 0),
    10: ('A', 1, 0),
    10.5: ('A', 1.5, 0),
    11: ('B', 0, 0),
    11.5: ('B', 0.5, 0)
}

#:
FORCESHARP = {
    0: ('B', 1, -1),
    0.5: ('B', 1.5, -1),
    1: ('B', 2, -1),
    1.5: ('C', 1.5, 0),
    2: ('C', 2, 0),
    2.5: ('D', 0.5, 0),
    3: ('D', 1, 0),
    3.5: ('D', 1.5, 0),
    4: ('D', 2, 0),
    4.5: ('E', 0.5, 0),
    5: ('E', 1, 0),
    5.5: ('E', 1.5, 0),
    6: ('E', 2, 0),
    6.5: ('F', 1.5, 0),
    7: ('F', 2, 0),
    7.5: ('G', 0.5, 0),
    8: ('G', 1, 0),
    8.5: ('G', 1.5, 0),
    9: ('G', 2, 0),
    9.5: ('A', 0.5, 0),
    10: ('A', 1, 0),
    10.5: ('A', 1.5, 0),
    11: ('A', 2, 0),
    11.5: ('B', 0.5, 0)
}

#:
FORCEFLAT = {
    0: ('D', -2, 0),
    0.5: ('D', -1.5, 0),
    1: ('D', -1, 0),
    1.5: ('D', -0.5, 0),
    2: ('E', -2, 0),
    2.5: ('E', -1.5, 0),
    3: ('F', -2, 0),
    3.5: ('F', -1.5, 0),
    4: ('F', -1, 0),
    4.5: ('F', -0.5, 0),
    5: ('G', -2, 0),
    5.5: ('G', -1.5, 0),
    6: ('G', -1, 0),
    6.5: ('G', -0.5, 0),
    7: ('A', -2, 0),
    7.5: ('A', -1.5, 0),
    8: ('A', -1, 0),
    8.5: ('A', -0.5, 0),
    9: ('B', -2, 0),
    9.5: ('B', -1.5, 0),
    10: ('C', -2, 1),
    10.5: ('C', -1.5, 1),
    11: ('C', -1, 1),
    11.5: ('C', -0.5, 1)
}

#:
SIGNS = {-2: 'flat-flat',
         -1.5: 'three-quarters-flat',
         -1: 'flat',
         -0.5: 'quarter-flat',
         0: 'natural',
         0.5: 'quarter-sharp',
         1: 'sharp',
         1.5: 'three-quarters-sharp',
         2: 'double-sharp'
         }


class Accidental(MusicTree, XMLWrapper):
    """
    Accidental is the class for managing :obj:`musicscore.midi.Midi`'s accidental sign and its pitch parameters: step, alter, octave.
    The parameter mode ('standard', 'enharmonic', 'flat', 'sharp', 'force-flat', 'force-sharp') can be used to set different enharmonic variants of the same
    pitch.
    """
    _ATTRIBUTES = {'mode', 'show', 'parent_midi'}

    XMLClass = XMLAccidental

    def __init__(self, mode='standard', show: Optional[bool] = None, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(value_='natural', **kwargs)
        self._mode = None
        self._show = None
        self.show = show
        self.mode = mode

    def _update(self):
        self._update_parent_midi()
        self._update_xml_object()

    def _update_parent_midi(self):
        if self.parent_midi and self.parent_midi.value != 0:
            self.parent_midi._update_pitch_parameters()

    def _update_xml_object(self):
        if self.sign:
            self._xml_object.value_ = self.sign

    @XMLWrapper.xml_object.getter
    def xml_object(self) -> Optional[XMLClass]:
        if self.parent_midi and self.parent_midi.value == 0:
            return None
        if self.show is True:
            return self._xml_object
        elif self.show is False:
            return None

    @property
    def mode(self):
        """
        permitted modes: ``standard``, ``flat``, ``sharp``, ``enharmonic``, ``force-sharp``, ``force-flat``

        :return: accidental mode which corresponds to global variables :obj:`~musicscore.accidental.STANDARD`, :obj:`~musicscore.accidental.FLAT`, :obj:`~musicscore.accidental.SHARP`,
                 :obj:`~musicscore.accidental.ENHARMONIC`, :obj:`~musicscore.accidental.FORCESHARP`, :obj:`~musicscore.accidental.FORCEFLAT`
        """
        return self._mode

    @mode.setter
    def mode(self, value):
        permitted = ('standard', 'flat', 'sharp', 'enharmonic', 'force-sharp', 'force-flat')
        if value not in permitted:
            raise TypeError(f'accidental_mode.value {value} must be in {permitted}')
        self._mode = value
        self._update_xml_object()
        self._update_parent_midi()

    @property
    def parent_midi(self) -> 'Midi':
        """
        :return: The midi parent of Accidental in the :obj:`~musicscore.musictree.MusicTree` structure. It is equivalent to :obj:`up` or :obj:`get_parent()`
        :rtype: :obj:`musicscore.midi.Midi`
        """
        return self.up

    @parent_midi.setter
    def parent_midi(self, val):
        val.add_child(self)

    @property
    def sign(self) -> Optional[str]:
        """
        Converts ``alter`` into sign depending on mode. ``parent_midi`` must be set first.

        :return: Possible values: ``flat-flat``, ``three-quarters-flat``, ``flat``, ``quarter-flat``, ``natural``,
                 ``quarter-sharp``, ``sharp``, ``three-quarters-sharp``, ``double-sharp``
        """
        try:
            alter = self.get_pitch_parameters()[1]
            return SIGNS[alter]
        except TypeError:
            return None

    @property
    def show(self) -> bool:
        """
        If ``False`` ``xml_object`` will be ``None`` and no accidental is shown in the :obj:`~musicscore.score.Score`.
        If ``True`` the accidental is shown. If the value of one of the following xml_object :obj:`~musicxml.xmlelement.xmlelement.XMLAccidental` properties:``self.parentheses``, ``self.editorial``.
        ``self.cautionary`` or ``self.bracket`` is set to yes, the accidental is always shown regardless of this
        property
        """
        if 'yes' in [self.parentheses, self.editorial, self.cautionary, self.bracket]:
            return True
        return self._show

    @show.setter
    def show(self, val):
        if val is not None and not isinstance(val, bool):
            raise TypeError
        if val != self._show:
            self._show = val
            try:
                self.up.up._update_xml_accidental()
            except AttributeError:
                pass

    def get_parent(self) -> 'Midi':
        """
        :return: parent
        :rtype: :obj:`~musicscore.midi.Midi`
        """
        return super().get_parent()

    def get_pitch_parameters(self, midi_value: Optional[float] = None) -> Optional[tuple]:
        """
        :return: A tuple consisting of pitch step name, alter value and octave value: ``(step, alter, octave)``

                If ``midi_value == 0`` (for a rest) return value is ``None``.

                If ``midi_value is None`` and a parent_midi exists, parent_midi's value will be used.
        """

        if midi_value is None:
            if self.parent_midi:
                midi_value = self.parent_midi.value
            else:
                return None
        if midi_value == 0:
            return None

        if self.mode == 'standard':
            output = STANDARD[midi_value % 12]

        elif self.mode == 'enharmonic':
            output = ENHARMONIC[midi_value % 12]

        elif self.mode == 'force-sharp':
            output = FORCESHARP[midi_value % 12]

        elif self.mode == 'force-flat':
            output = FORCEFLAT[midi_value % 12]

        elif self.mode == 'flat':
            output = FLAT[midi_value % 12]

        elif self.mode == 'sharp':
            output = SHARP[midi_value % 12]

        else:
            raise ValueError
        return output[0], output[1], output[2] + (int(midi_value // 12)) - 1

    def __copy__(self):
        return self.__class__(mode=self.mode, show=self.show)



from pathlib import Path

from musicscore import Score, C, D, Chord, Time
from musicscore.tests.util import IdTestCase

"""Accidentals can be cautionary or editorial. Each measure has a normal accidental, an editorial, a cautionary and 
an editorial and cautionary accidental."""


class TestLily01e(IdTestCase):
    def test_lily_01e_Pitches_ParenthesizedAccidentals(self):
        score = Score('cautionary accidentals')
        p = score.add_part('cautionary')
        p.add_measure(Time(5, 4))
        midis = [C(4, 'bb') for _ in range(5)] + \
                [C(4, 'three-quarters-flat') for _ in range(5)] + \
                [C(4, 'b') for _ in range(5)] + \
                [C(4, 'quarter-flat') for _ in range(5)] + \
                [C(4, 'natural') for _ in range(5)] + \
                [C(4, 'quarter-sharp') for _ in range(5)] + \
                [C(4, '#') for _ in range(5)] + \
                [C(4, 'three-quarters-sharp') for _ in range(5)] + \
                [C(4, '##') for _ in range(5)]
        for index, midi in enumerate(midis):
            if (index + 1) % 5 == 0:
                midi.accidental.bracket = 'yes'
                ch = Chord(midi, 1)
                ch.add_lyric('bracket')
            elif (index + 1) % 5 == 1:
                ch = Chord(midi, 1)
            elif (index + 1) % 5 == 2:
                midi.accidental.parentheses = 'yes'
                ch = Chord(midi, 1)
                ch.add_lyric('parentheses')
            elif (index + 1) % 5 == 3:
                midi.accidental.cautionary = 'yes'
                ch = Chord(midi, 1)
                ch.add_lyric('cautionary')
            else:
                midi.accidental.editorial = 'yes'
                ch = Chord(midi, 1)
                ch.add_lyric('editorial')
            p.add_chord(ch)

        xml_file = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_file)



from pathlib import Path

from musicscore.accidental import Accidental
from musicscore.chord import Chord
from musicscore.measure import generate_measures
from musicscore.midi import Midi
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import IdTestCase
from musicscore.time import Time


class TestHelloChromatics(IdTestCase):
    def test_chromatic_up_quarter_grid(self):
        """
        Scale chromatically up with different quarter_durations and time signatures
        """
        s = Score()
        p = s.add_child(Part('p1'))
        quarter_durations = list(range(10, 0, -1))
        """
        Adding measures
        """
        measures = generate_measures(times=[Time(qd, 4) for qd in quarter_durations])
        for m in measures:
            p.add_child(m)
        """
        Adding chords
        """
        min_midi = 60
        for m, qd in zip(measures, quarter_durations):
            m._add_chord(Chord(quarter_duration=qd, midis=Midi(min_midi + 10 - qd, accidental=Accidental(mode='sharp'))))
        """
        Exporting
        """
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



"""
All <direction> elements defined in MusicXML.
"""
from pathlib import Path

from musicscore.tests.util import IdTestCase
from musicxml.xmlelement.xmlelement import XMLRehearsal, XMLDynamics, XMLOtherDynamics, XMLSegno, XMLCoda, XMLWords, \
    XMLEyeglasses, XMLDashes, XMLBracket, XMLOctaveShift, XMLPedal, XMLMetronome, XMLBeatUnit, XMLPerMinute, \
    XMLHarpPedals, XMLPedalStep, XMLPedalAlter, XMLPedalTuning, XMLDamp, XMLDampAll, XMLScordatura, XMLAccord, \
    XMLTuningStep, XMLTuningOctave, XMLTuningAlter, XMLAccordionRegistration, XMLAccordionHigh, XMLAccordionMiddle, \
    XMLAccordionLow, XMLWedge, XMLPrint, XMLBarline, XMLBarStyle

""""
XML_DIRECTION_TYPE_CLASSES = [
    XMLSymbol, XMLStringMute, XMLImage, XMLPrincipalVoice, XMLPercussion, XMLStaffDivide, XMLOtherDirection
]
"""
from musicscore import Score, Chord


class TestLily31a(IdTestCase):
    def test_lily_31a_Directions(self):

        score = Score(title="MusicXML directions (attached to staff)")
        part = score.add_part('p1')

        for rehearsal, enclosure in zip(['A', 'B', 'Test', 'Crc'], ['square', 'none', 'rectangle', 'circle']):
            r = XMLRehearsal(rehearsal, enclosure=enclosure)
            ch = Chord(60, 1)
            ch.add_direction_type(r)
            part.add_chord(ch)

        for dt in [XMLSegno(), XMLCoda(), XMLWords('words'), XMLEyeglasses()]:
            ch = Chord(60, 1)
            if isinstance(dt, XMLWords):
                ch.add_direction_type(dt, 'below')
            else:
                ch.add_direction_type(dt, 'above')
            part.add_chord(ch)

        dynamics = ['p', 'pp', 'ppp', 'pppp', 'ppppp', 'pppppp', 'f', 'ff', 'fff', 'ffff', 'fffff', 'ffffff', 'mp',
                    'mf',
                    'sf', 'sfp', 'sfpp', 'fp', 'rf', 'rfz', 'sfz', 'sffz', 'fz']
        for d in dynamics:
            ch = Chord(60, 1)
            ch.add_dynamics(d)
            part.add_chord(ch)

        ch = Chord(60, 1)
        d = XMLDynamics()
        d.add_child(XMLOtherDynamics('abc-ffz'))
        ch.add_direction_type(d, placement='below')
        part.add_chord(ch)

        for w in ['crescendo', 'stop']:
            ch = Chord(60, 1)
            ch.add_wedge(w)
            part.add_chord(ch)

        for x in ['start', 'stop']:
            ch = Chord(60, 1)
            ch.add_direction_type(XMLDashes(type=x))
            part.add_chord(ch)

        for x in ['start', 'stop']:
            ch = Chord(60, 1)
            ch.add_direction_type(XMLBracket(type=x, line_end='down'))
            part.add_chord(ch)

        for x in ['down', 'stop']:
            ch = Chord(60, 1)
            ch.add_direction_type(XMLOctaveShift(type=x))
            part.add_chord(ch)

        for x in ['start', ('stop', 'start'), 'continue', 'stop']:
            ch = Chord(60, 1)
            if isinstance(x, tuple):
                ch.add_direction_type(XMLPedal(type=x[0], relative_x=-25, relative_y=-10), 'below')
                ch.add_direction_type(XMLPedal(type=x[1], relative_y=-10), 'below')
            else:
                ch.add_direction_type(XMLPedal(type=x, relative_y=-10), 'below')
            part.add_chord(ch)

        ch = Chord(60, 1)
        m = XMLMetronome()
        m.add_child(XMLBeatUnit('quarter'))
        m.add_child(XMLPerMinute('60'))
        ch.add_direction_type(m)
        part.add_chord(ch)

        ch = Chord(60, 1)
        steps = ['D', 'C', 'B', 'E', 'F', 'G', 'A']
        alters = [0, -1, -1, 0, 0, 1, -1]
        hp = XMLHarpPedals(relative_y=-20, font_size=14)
        for s, a in zip(steps, alters):
            pt = hp.add_child(XMLPedalTuning())
            pt.add_child(XMLPedalStep(s))
            pt.add_child(XMLPedalAlter(a))
        ch.add_direction_type(hp, 'below')
        part.add_chord(ch)

        ch = Chord(60, 1)
        ch.add_direction_type(XMLDamp())
        part.add_chord(ch)

        ch = Chord(60, 1)
        ch.add_direction_type(XMLDampAll())
        part.add_chord(ch)

        """
        <scordatura>
           <accord string="1">
              <tuning-step>E</tuning-step>
              <tuning-octave>5</tuning-octave>
           </accord>
           <accord string="2">
              <tuning-step>A</tuning-step>
              <tuning-octave>4</tuning-octave>
           </accord>
           <accord string="3">
              <tuning-step>E</tuning-step>
              <tuning-alter>-1</tuning-alter>
              <tuning-octave>4</tuning-octave>
           </accord>
           <accord string="4">
              <tuning-step>A</tuning-step>
              <tuning-octave>3</tuning-octave>
           </accord>
        </scordatura>
        """

        sc = XMLScordatura()
        steps = ['E', 'A', 'E', 'A']
        alters = [None, None, -1, None]
        octaves = [5, 4, 4, 3]
        for index, (s, a, o) in enumerate(zip(steps, alters, octaves)):
            st = sc.add_child(XMLAccord(string=index + 1))
            st.add_child(XMLTuningStep(s))
            if a:
                st.add_child(XMLTuningAlter(a))
            st.add_child(XMLTuningOctave(o))

        ch = Chord(60, 1)
        ch.add_direction_type(sc)
        ch.add_lyric('Scordatura')
        part.add_chord(ch)

        """
        <direction>
           <direction-type>
              <accordion-registration>
                 <accordion-high/>
                 <accordion-middle>3</accordion-middle>
              </accordion-registration>
           </direction-type>
        </direction>
        """
        ch = Chord(60, 1)
        ag = XMLAccordionRegistration()
        ag.add_child(XMLAccordionHigh())
        ag.add_child(XMLAccordionMiddle(2))
        ag.add_child(XMLAccordionLow())
        ch.add_direction_type(ag)
        part.add_chord(ch)

        ch = Chord(0, 2)
        part.add_chord(ch)

        ch = Chord(60, 1)
        ch.add_dynamics('p')
        ch.add_direction_type(XMLWords('subito', font_style='italic', relative_y=-20), placement='below')
        part.add_chord(ch)

        ch = Chord(60, 1)
        ch.add_dynamics('ppp')
        ch.add_wedge(XMLWedge(type='crescendo', relative_x=25))
        part.add_chord(ch)

        ch = Chord(60, 1)
        ch.add_dynamics('fff')
        ch.add_wedge(XMLWedge(type='stop', relative_x=-5))
        part.add_chord(ch)

        ch = Chord(0, 1)
        part.add_chord(ch)

        new_systems = [2, 5, 9, 12]
        for m in new_systems:
            part.get_measure(m).xml_object.add_child(XMLPrint(new_system='yes'))

        barline = XMLBarline(location='right')
        barline.add_child(XMLBarStyle('light-light'))
        part.get_measure(13).xml_object.add_child(barline)

        barline = XMLBarline(location='right')
        barline.add_child(XMLBarStyle('light-heavy'))
        part.get_measure(14).xml_object.add_child(barline)

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



"""
All different types of glissando defined in MusicXML
"""
from pathlib import Path

from musicscore import Score, G, Chord, F
from musicscore.tests.util import IdTestCase
from musicxml import XMLGlissando, XMLSlide

types = ['normal glissando', 'solid (+text)', 'dashed', 'dotted', 'wavy', 'normal slide', 'solid (+text)', 'dashed',
         'dotted', 'wavy']


class TestLily33h(IdTestCase):
    def test_lily_33h_Spanners_Glissando(self):
        score = Score()
        part = score.add_part('p1')
        line_types = [None, 'solid', 'dashed', 'dotted', 'wavy']
        for i, t in enumerate(types):
            chords = [Chord(G(4), 1), Chord(F(5), 1)]
            texts = t.split(' ')
            chords[0].add_lyric(texts[0])
            if len(texts) > 1:
                chords[1].add_lyric(texts[1])

            if i < 5:
                chords[0].add_x(XMLGlissando(type='start', line_type=line_types[i]))
                chords[1].add_x(XMLGlissando(type='stop'))
            else:
                chords[0].add_x(XMLSlide(type='start', line_type=line_types[i-5]))
                chords[1].add_x(XMLSlide(type='stop'))
            [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from typing import Optional, Set, List

from musicscore.finalize import FinalizeMixin
from musicscore.quantize import QuantizeMixin
from musicxml.xmlelement.xmlelement import XMLStaff

from musicscore.clef import Clef
from musicscore.exceptions import StaffHasNoParentError, AlreadyFinalizedError
from musicscore.musictree import MusicTree
from musicscore.voice import Voice
from musicscore.xmlwrapper import XMLWrapper

__all__ = ['Staff']


class Staff(MusicTree, QuantizeMixin, FinalizeMixin, XMLWrapper):
    _ATTRIBUTES = {'clef', 'default_clef', 'number'}
    _ATTRIBUTES = _ATTRIBUTES.union(MusicTree._ATTRIBUTES)
    _ATTRIBUTES = _ATTRIBUTES.union(QuantizeMixin._ATTRIBUTES)
    XMLClass = XMLStaff

    def __init__(self, number=None, clef=None, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(value_=1, **kwargs)
        self._number = None
        self._clef = None
        self.clef = clef
        self.number = number

    @property
    def clef(self) -> Clef:
        """
        :type: :obj:`~musicscore.clef.Clef`
        :return: Clef associated with the staff.
        :rtype: :obj:`~musicscore.clef.Clef`
        """
        if not self._clef:
            if self.get_previous_staff() and self.get_previous_staff().clef:
                self._clef = self.get_previous_staff().clef.__copy__()
                self._clef.show = False
        return self._clef

    @clef.setter
    def clef(self, val):
        if val is not None and not isinstance(val, Clef):
            raise TypeError
        if val and not val.number and self.clef:
            val.number = self.clef.number
        self._clef = val

    @property
    def number(self):
        """
        :type: None or int. If None number is set to 1.
        :return: positive int or None
        """
        if self._number is not None:
            return self.xml_object.value_
        else:
            return self._number

    @number.setter
    def number(self, val):
        self._number = val
        if val is None:
            self.xml_object.value_ = 1
        else:
            self.xml_object.value_ = val

    @XMLWrapper.xml_object.getter
    def xml_object(self) -> XMLClass:
        return super().xml_object

    def add_child(self, child: Voice) -> Voice:
        """
        - Adds a :obj:`~musicscore.voice.Voice` as child to staff.
        - If voice number is ``None``, it is determined as length of children + 1.
        - If voice number is already set an is not equal to length of children + 1. a ``ValueError`` is raised.
        - If Staff has no parent :obj:`~musicscore.measure.Measure` ``StaffHasNoParentError`` is raised.
        - After adding voice its :obj:`~musicscore.voice.Voice.update_beats()` is called.

        :param child: :obj:`~musicscore.voice.Voice` , required
        :return: added :obj:`~musicscore.voice.Voice`
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_child')
        self._check_child_to_be_added(child)

        if not self.up:
            raise StaffHasNoParentError('A child Voice can only be added to a Staff if staff has a Measure parent.')

        if child.number is not None and child.number != len(self.get_children()) + 1:
            raise ValueError(f'Voice number must be None or {len(self.get_children()) + 1}')
        if child.number is None:
            child.number = len(self.get_children()) + 1
        else:
            child.number = len(self.get_children()) + 1

        child._parent = self
        self._children.append(child)

        return child

    def add_voice(self, voice_number: Optional[int] = None) -> Voice:
        """
        - Creates and adds a new :obj:`~musicscore.voice.Voice` object as child to staff.
        - If voice number is greater than length of children + 1 all missing voices are created and added first.

        :param voice_number: positive int or None. If ``None`` voice number is determined as length of children + 1.
        :return: new :obj:`~musicscore.voice.Voice`
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_voice')
        if voice_number is None:
            voice_number = len(self.get_children()) + 1

        voice_object = self.get_voice(voice_number=voice_number)

        if voice_object is None:
            for _ in range(voice_number - len(self.get_children())):
                voice_object = self.add_child(Voice())
                voice_object.update_beats()
            return voice_object

        voice_object.update_beats()
        return voice_object

    def get_children(self) -> List[Voice]:
        """
        :return: list of added children.
        :rtype: List[:obj:`~musicscore.voice.Voice`]
        """
        return super().get_children()

    def get_parent(self) -> 'Measure':
        """
        :return: parent
        :rtype: :obj:`~musicscore.measure.Measure`
        """
        return super().get_parent()

    def get_previous_staff(self) -> Optional['Staff']:
        """
        :return: :obj:`Staff` with the same number in previous :obj:`~musicscore.measure.Measure` if existing else ``None``
        """
        if self.up and self.up.previous:
            my_index = self.up.get_children().index(self)
            try:
                return self.up.previous.get_children()[my_index]
            except IndexError:
                return None
        return None

    def get_last_pitch_steps_with_accidentals(self) -> Set[str]:
        """
        :return: a set of pitch steps with not natural accidental. This method is used to take track of accidental signs which needs to
                 be shown hidden
        :rtype:  Set[str]
        """
        output = set()
        for v in self.get_children():
            if v.get_chords():
                last_chord = v.get_chords()[-1]
                if not last_chord.is_rest:
                    for m in last_chord.midis:
                        if m.accidental.sign != 'natural':
                            step = m.accidental.get_pitch_parameters()[0]
                            if step not in output:
                                output.add(step)
        return output



import inspect

from musicscore import C, B, G, Part, Chord, Score, Time
from musicscore.tests.util import TestCase, generate_path, IdTestCase
from musicxml.xmlelement.xmlelement import XMLNotehead


class TestMidiNotes(TestCase):
    def test_midi_note(self):
        m = C(4, 's')
        assert m.accidental.get_pitch_parameters() == ('C', 1, 4)
        m = C(0)
        assert m.accidental.get_pitch_parameters() == ('C', 0, 0)
        with self.assertRaises(ValueError):
            B(-1)
        with self.assertRaises(ValueError):
            G(9, 's')

    def test_midi_note_half_tone_accidentals(self):
        signs = ['double-flat', 'flat-flat', 'ff', 'bb', 'flat', 'f', 'b', 'sharp', 's', '#', 'double-sharp',
                 'sharp-sharp',
                 'ss', '##']
        for sign in signs:
            m = C(4, sign)
            if sign in ['double-flat', 'flat-flat', 'ff', 'bb']:
                assert m.value == 58
                assert m.accidental.get_pitch_parameters() == ('C', -2, 4)
                assert m.accidental.sign == 'flat-flat'

            elif sign in ['flat', 'f', 'b']:
                assert m.value == 59
                assert m.accidental.sign == 'flat'
                assert m.accidental.get_pitch_parameters() == ('C', -1, 4)
            elif sign in ['sharp', 's', '#']:
                assert m.value == 61
                assert m.accidental.sign == 'sharp'
                assert m.accidental.get_pitch_parameters() == ('C', 1, 4)
            elif sign in ['double-sharp', 'sharp-sharp', 'ss', '##']:
                assert m.value == 62
                assert m.accidental.sign == 'double-sharp'
                assert m.accidental.get_pitch_parameters() == ('C', 2, 4)

    def test_midi_note_quarter_tones(self):
        m1 = C(4, 'three-quarters-flat')
        assert m1.value == 60 - 3 / 2
        assert m1.accidental.sign == 'three-quarters-flat'
        assert m1.accidental.get_pitch_parameters() == ('C', -1.5, 4)
        m2 = C(4, 'quarter-flat')
        m3 = C(4, 'quarter-sharp')
        m4 = C(4, 'three-quarters-sharp')

    def test_midi_note_copy(self):
        m = C(4, '#')
        m.add_tie('start')
        copied = m.__copy__()
        assert m != copied
        assert m.value == copied.value
        assert id(m.accidental) == id(copied.accidental)
        assert id(m._ties) == id(copied._ties)

        copied = m.__deepcopy__()
        assert m != copied
        assert m.value == copied.value
        assert m.accidental != copied.accidental
        assert m.accidental.mode == copied.accidental.mode
        assert m.accidental.show == copied.accidental.show
        assert id(m._ties) != id(copied._ties)
        assert m._ties == copied._ties

        copied = m._copy_for_split()
        assert m != copied
        assert m.value == copied.value
        assert m.accidental != copied.accidental
        assert m.accidental.mode == copied.accidental.mode
        assert m.accidental.show == copied.accidental.show
        assert id(m._ties) != id(copied._ties)
        assert m._ties != copied._ties

        #
        # copied = m.__deepcopy__()
        # assert m != copied
        # assert m.value == copied.value
        # assert m.accidental != copied.accidental
        # assert m.accidental.mode == copied.accidental.mode
        # assert m.accidental.show == copied.accidental.show





class TestMidiNoteNoteHead(IdTestCase):
    def test_notehead_property(self):
        m = C(4)
        m.notehead = 'square'
        assert isinstance(m.notehead, XMLNotehead)
        assert m.notehead.value_ == 'square'

    def test_notehead_after_finalize(self):
        p = Part('p1')
        ch = Chord(C(4), 1)
        ch.midis[0].notehead = 'square'
        p.add_chord(ch)
        p.finalize()
        assert ch.midis[0].parent_note.xml_notehead.value_ == 'square'

    def test_notehead_copy_for_split(self):
        midi = C(4)
        midi.notehead = 'square'
        copied = midi._copy_for_split()
        assert copied.notehead.value_ == 'square'

    def test_midi_note_notehead_after_split(self):
        s = Score()
        p = s.add_part('p1')
        ch = Chord(C(4), 3)
        ch.midis[0].notehead = 'square'
        p.add_measure(Time(2, 4))
        p.add_chord(ch)
        path = generate_path(inspect.currentframe())
        s.export_xml(path)
        assert p.get_chords()[-1].midis[0].notehead.value_ == 'square'
        assert p.get_chords()[-1].midis[0].parent_note.xml_notehead.value_ == 'square'


from musicscore import Part
from musicscore.chord import GraceChord
from musicscore.exceptions import ChordException
from musicscore.tests.util import IdTestCase
from musicxml.xmlelement.xmlelement import XMLType


class TestGraceChord(IdTestCase):
    def test_grace_chord_init(self):
        gch = GraceChord()
        assert gch.quarter_duration == 0
        gch.quarter_duration = 0
        with self.assertRaises(ChordException):
            gch.quarter_duration = 2

    def test_grace_chord_type(self):
        p = Part('p1')
        gch = GraceChord(60)

        gch.type_ = '16th'
        assert isinstance(gch.type_, XMLType)
        gch.type_ = XMLType('eighth')
        assert isinstance(gch.type_, XMLType)
        gch.type_ = None
        assert gch.type_ is None

        gch.type_ = '16th'
        p.add_chord(gch)
        p.finalize()
        assert p.get_chords()[0].notes[0].xml_type.value_ == '16th'



import math
from typing import Union, List

from musicscore.exceptions import WrongNumberOfChordsError, LyricSyllabicOrExtensionError
from musicxml.xmlelement.xmlelement import *

note_types = {(1, 12): '32nd',
              (1, 11): '32nd',
              (2, 11): '16th',
              (3, 11): '16th',
              (4, 11): 'eighth',
              (6, 11): 'eighth',
              (8, 11): 'quarter',
              (1, 10): '32nd',
              (3, 10): '16th',
              (1, 9): '32nd',
              (2, 9): '16th',
              (4, 9): 'eighth',
              (8, 9): 'quarter',
              (1, 8): '32nd',
              (3, 8): '16th',
              (7, 8): 'eighth',
              (1, 7): '16th',
              (2, 7): 'eighth',
              (3, 7): 'eighth',
              (4, 7): 'quarter',
              (6, 7): 'quarter',
              (1, 6): '16th',
              (1, 5): '16th',
              (2, 5): 'eighth',
              (3, 5): 'eighth',
              (4, 5): 'quarter',
              (6, 5): 'quarter',
              (8, 5): 'half',
              (1, 4): '16th',
              (2, 4): 'eighth',
              (3, 4): 'eighth',
              (7, 4): 'quarter',
              (4, 3): 'half',
              (1, 3): 'eighth',
              (2, 3): 'quarter',
              (3, 2): 'quarter',
              (1, 2): 'eighth',
              (1, 1): 'quarter',
              (2, 1): 'half',
              (3, 1): 'half',
              (4, 1): 'whole',
              (6, 1): 'whole',
              (8, 1): 'breve',
              (12, 1): 'breve'
              }

XML_ARTICULATION_CLASSES = [XMLAccent, XMLStrongAccent, XMLStaccato, XMLTenuto, XMLDetachedLegato, XMLStaccatissimo,
                            XMLSpiccato, XMLScoop, XMLPlop, XMLDoit, XMLFalloff, XMLBreathMark, XMLCaesura, XMLStress,
                            XMLUnstress]

XML_TECHNICAL_CLASSES = [XMLUpBow, XMLDownBow, XMLHarmonic, XMLOpenString, XMLThumbPosition, XMLFingering, XMLPluck,
                         XMLDoubleTongue,
                         XMLTripleTongue, XMLStopped, XMLSnapPizzicato, XMLFret, XMLString, XMLHammerOn, XMLPullOff,
                         XMLBend, XMLTap,
                         XMLHeel, XMLToe, XMLFingernails, XMLHole, XMLArrow, XMLHandbell, XMLBrassBend, XMLFlip,
                         XMLSmear, XMLOpen,
                         XMLHalfMuted, XMLHarmonMute, XMLGolpe, XMLOtherTechnical]

XML_ORNAMENT_CLASSES = [XMLDelayedInvertedTurn, XMLDelayedTurn, XMLHaydn, XMLInvertedMordent,
                        XMLInvertedTurn,
                        XMLInvertedVerticalTurn, XMLMordent, XMLOtherOrnament, XMLSchleifer, XMLShake, XMLTremolo,
                        XMLTrillMark, XMLTurn,
                        XMLVerticalTurn, XMLWavyLine]

XML_DYNAMIC_CLASSES = [XMLF, XMLFf, XMLFff, XMLFfff, XMLFffff, XMLFfffff, XMLFp, XMLFz, XMLMf, XMLMp, XMLP, XMLPf,
                       XMLPp, XMLPpp, XMLPppp,
                       XMLPpppp, XMLPppppp, XMLRf, XMLRfz, XMLSf, XMLSffz, XMLSfp, XMLSfpp, XMLSfz, XMLSfzp,
                       XMLOtherDynamics]

XML_OTHER_NOTATIONS = [XMLArpeggiate, XMLFermata, XMLFootnote, XMLGlissando, XMLLevel, XMLNonArpeggiate,
                       XMLOtherNotation, XMLSlide,
                       XMLSlur]

# XML_DIRECTION_TYPE_CLASSES = [
#     XMLRehearsal, XMLSegno, XMLCoda, XMLWords, XMLSymbol, XMLWedge, XMLDashes, XMLBracket, XMLPedal,
#     XMLMetronome, XMLOctaveShift, XMLHarpPedals, XMLDamp, XMLDampAll, XMLEyeglasses, XMLStringMute, XMLScordatura,
#     XMLImage, XMLPrincipalVoice, XMLPercussion, XMLAccordionRegistration, XMLStaffDivide, XMLOtherDirection
# ]

XML_DIRECTION_TYPE_CLASSES = [
    XMLRehearsal, XMLSegno, XMLCoda, XMLWords, XMLSymbol, XMLWedge, XMLDashes, XMLBracket, XMLPedal,
    XMLMetronome, XMLOctaveShift, XMLHarpPedals, XMLDamp, XMLDampAll, XMLEyeglasses, XMLStringMute, XMLScordatura,
    XMLPrincipalVoice, XMLPercussion, XMLAccordionRegistration, XMLStaffDivide, XMLOtherDirection
]

XML_ORNAMENT_AND_OTHER_NOTATIONS = [XMLAccidentalMark]

XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS = [XMLDynamics]


def lcm(l):
    return math.lcm(*l)


def dToX(input_list, first_element=0):
    if isinstance(input_list, list) is False:
        raise TypeError('xToD(input_list)')
    else:
        output = [first_element]
        for i in range(len(input_list)):
            output.append(input_list[i] + output[i])
        return output


def xToD(input_list):
    result = []
    for i in range(1, len(input_list)):
        result.append(input_list[i] - input_list[i - 1])
    return result


def isinstance_as_string(child: object, parent_class_names: Union[str, List[str]]) -> bool:
    """
    This function can be used to check if some class names (parent_class_names) can be found in another class's __mro__.
    If parent classes cannot be imported due to recursive imports this can be used instead of isinstance function.
    :param object child:
    :param str/[str] parent_class_names:
    :return: bool
    """
    if isinstance(parent_class_names, str):
        parent_class_names = [parent_class_names]

    for parent_class_name in parent_class_names:
        if parent_class_name not in [cls.__name__ for cls in child.__class__.__mro__]:
            return False
    return True


def chord_is_in_a_repetition(chord):
    my_index = chord.up.up.get_chords().index(chord)
    if my_index > 0 and not chord.is_tied_to_previous:
        all_previous_chords = [chord.up.up.get_chords()[my_index - i] for i in range(1, my_index + 1)]
        if set([ch.is_tied_to_previous for ch in all_previous_chords]) == {True}:
            return False
        previous_chord = all_previous_chords[0]
        if chord.has_same_pitches(previous_chord):
            return True
    return False


def slur_chords(chords, number=1, **kwargs):
    if len(chords) < 2:
        raise WrongNumberOfChordsError('util.slur_chords needs at list two chords.')

    chords[0].add_x(XMLSlur(type='start', number=number, **kwargs))
    chords[-1].add_x(XMLSlur(type='stop', number=number))
    for ch in chords[1:-1]:
        ch.add_x(XMLSlur(type='continue', number=number))


def trill_chords(chords, number=1, placement='above', **kwargs):
    if len(chords) < 2:
        raise WrongNumberOfChordsError('util.trill_chords needs at list two chords.')
    chords[0].add_x(XMLTrillMark(placement=placement, **kwargs))
    chords[0].add_x(XMLWavyLine(type='start', number=number, **kwargs), placement=placement)
    chords[-1].add_x(XMLWavyLine(type='stop', number=number, **kwargs), placement=placement)
    for ch in chords[1:-1]:
        ch.add_x(XMLWavyLine(type='continue', number=number, **kwargs), placement=placement)


def wedge_chords(chords, wedge_type, number=1, placement='below', **kwargs):
    if len(chords) < 2:
        raise WrongNumberOfChordsError('util.wedge_chords needs at list two chords.')

    chords[0].add_x(XMLWedge(type=wedge_type, number=number, **kwargs), placement=placement)
    chords[-1].add_x(XMLWedge(type='stop', number=number), placement=placement)
    for ch in chords[1:-1]:
        ch.add_x(XMLWedge(type='continue', number=number), placement=placement)


def bracket_chords(chords, line_type='solid', start_line_end='down', end_line_end='down', placement='above', number=1):
    if len(chords) < 2:
        raise WrongNumberOfChordsError('util.bracket_chords needs at list two chords.')

    chords[0].add_x(XMLBracket(type='start', line_end=start_line_end, line_type=line_type, number=number),
                    placement=placement)
    chords[-1].add_x(XMLBracket(type='stop', line_end=end_line_end, line_type=line_type, number=number),
                     placement=placement)
    for ch in chords[1:-1]:
        ch.add_x(XMLBracket(type='continue', line_end='none', line_type=line_type, number=number), placement=placement)


def octave_chords(chords, type='down', size=8, number=1):
    try:
        len(chords)
    except TypeError:
        chords = [chords]
    if type == 'down':
        placement = 'above'
    else:
        placement = 'below'
    # if len(chords) < 2:
    #     raise WrongNumberOfChordsError('util.octave_chords needs at list two chords.')

    chords[0].add_x(XMLOctaveShift(type=type, size=size, number=number), placement=placement)
    chords[-1].add_x(XMLOctaveShift(type='stop', size=size, number=number), placement=placement)
    for ch in chords[1:-1]:
        ch.add_x(XMLOctaveShift(type='continue', size=size, number=number), placement=placement)


def _generate_lyrics(lyrics, number=1, show_number=False, mode='list', **kwargs):
    def _get_syllables_extensions_from_group(syllabic_group):
        if syllabic_group[0] is None:
            raise LyricSyllabicOrExtensionError(
                f'Syllabic or extension cannot be added to the beginning of a syllabic group.')
        extensions_ = []
        syllables_ = list(syllabic_group)[:]
        while syllables_[-1] is None:
            extensions_.append(syllables_.pop())
        return [syllables_, extensions_]

    if isinstance(lyrics, str) or isinstance(lyrics, tuple):
        lyrics = [lyrics]

    output = []
    if mode == 'list':
        for i in range(len(lyrics)):
            ll = lyrics[i]
            if isinstance(ll, str):
                xl = XMLLyric(number=str(number), **kwargs)
                xl.xml_text = ll
                xl.xml_syllabic = 'single'
                output.append(xl)

            elif isinstance(ll, tuple) or isinstance(ll, list):
                syllables, extensions = _get_syllables_extensions_from_group(ll)
                if len(syllables) == 1:
                    xl = XMLLyric(number=str(number), **kwargs)
                    xl.xml_text = syllables[0]
                    xl.xml_syllabic = 'single'
                    if len(extensions) > 0:
                        xl.xml_extend = XMLExtend(type='start')

                    output.append(xl)
                else:
                    for j in range(len(syllables)):
                        xl = XMLLyric(number=str(number), **kwargs)
                        l = syllables[j]
                        if l:
                            xl.xml_text = l
                            if j == 0:
                                xl.xml_syllabic = 'begin'
                            elif j == len(syllables) - 1:
                                xl.xml_syllabic = 'end'
                                if len(extensions) > 0:
                                    xl.xml_extend = XMLExtend(type='start')
                            else:
                                xl.xml_syllabic = 'middle'
                        else:
                            xl = None
                        output.append(xl)

                for k in range(len(extensions)):
                    xl = XMLLyric(number=str(number), **kwargs)
                    if k == len(extensions) - 1:
                        xl.xml_extend = XMLExtend(type='stop')
                    else:
                        xl.xml_extend = XMLExtend(type='continue')
                    xl.xsd_check = False
                    output.append(xl)
            elif ll is None:
                output.append(None)
            else:
                raise NotImplementedError(ll)
    else:
        raise NotImplementedError
    if show_number:
        first = XMLLyric(**kwargs)
        for k, i in output[0].attributes.items():
            setattr(first, k, i)
        first.add_child(XMLSyllabic('single'))
        first.add_child(XMLText(f'{number}.'))
        first.add_child(XMLElision(' '))

        for child in output[0].get_children():
            first.add_child(child)

        output[0] = first
    return output



from quicktions import Fraction
from unittest import TestCase

from musicscore.beat import Beat
from musicscore.exceptions import QuarterDurationIsNotWritable
from musicscore.quarterduration import QuarterDuration, _check_quarter_duration_value


class TestQuarterDuration(TestCase):
    def test_quarter_duration_init(self):
        qd = QuarterDuration(3, 2)
        assert (qd.numerator, qd.denominator) == (3, 2)
        assert qd == 1.5
        assert qd == QuarterDuration(3, 2)
        qd = QuarterDuration(1.2)
        assert qd.as_integer_ratio() == (6, 5)
        assert qd == QuarterDuration(6, 5)
        assert qd == 1.2

    def test_quarter_duration_operands(self):
        qd = QuarterDuration(2)
        assert qd == 2
        assert qd <= 2
        assert qd >= 2
        assert not qd == 3
        assert qd < 3
        assert qd != 3
        assert qd > 1
        qd = QuarterDuration(1, 6)
        assert qd == Fraction(1, 6)
        assert qd == 1 / 6
        assert qd <= 1 / 6
        assert qd >= 1 / 6
        assert QuarterDuration(1, 3) == 1 / 3
        assert QuarterDuration(6 / 5) == 6 / 5
        assert QuarterDuration(6, 5) == 1.2
        assert QuarterDuration(1.2) == QuarterDuration(6, 5)
        assert QuarterDuration(Fraction(6 / 5)) == 6 / 5
        assert [QuarterDuration(7, 20), [QuarterDuration(3, 2), QuarterDuration(1)], QuarterDuration(23, 20)] == [0.35,
                                                                                                                  [1.5,
                                                                                                                   1],
                                                                                                                  1.15]
        assert [QuarterDuration(0.35), [QuarterDuration(1.5), QuarterDuration(1)], QuarterDuration(1.15)] == [0.35,
                                                                                                              [1.5, 1],
                                                                                                              1.15]

    def test_add_etc(self):
        assert isinstance(1 / 6 + QuarterDuration(1, 6), QuarterDuration)
        assert QuarterDuration(3, 4) + 0.25 == 1
        assert 1 / 6 + QuarterDuration(1, 6) == QuarterDuration(1, 3)
        assert QuarterDuration(1, 6) + 1 / 6 == QuarterDuration(1, 3)

        assert 1 / 6 + QuarterDuration(1, 6) == 1 / 3
        assert QuarterDuration(1, 6) + 1 / 6 == 1 / 3
        assert QuarterDuration(1, 6) + QuarterDuration(1, 6) == QuarterDuration(1, 3)
        assert QuarterDuration(1, 6) + QuarterDuration(1, 6) == 1 / 3
        assert QuarterDuration(1, 6) * 2 == 1 / 3
        assert 2 * QuarterDuration(1, 6) == 1 / 3

    def test_rtruediv(self):
        assert 2 / (1 / 2) == 4
        assert 2 / QuarterDuration(1, 2) == 4

    def test_split_quarter_duration(self):
        """
        Test if get_beatwise_sections can split quarter duration to writable sections.
        """
        assert QuarterDuration(4).get_beatwise_sections(beats=[Beat(1), Beat(1), Beat(1), Beat(1)]) == [[4], None]
        assert QuarterDuration(3.5).get_beatwise_sections(beats=[Beat(1), Beat(1), Beat(1), Beat(1)]) == [[3, 0.5],
                                                                                                          None]
        assert QuarterDuration(3.5).get_beatwise_sections(offset=0.5, beats=[Beat(1), Beat(1), Beat(1), Beat(1)]) == [
            [0.5, 3], None]
        assert QuarterDuration(4).get_beatwise_sections(offset=0.5, beats=[Beat(1), Beat(1), Beat(1), Beat(1)]) == [
            [0.5, 3], 0.5]
        assert QuarterDuration(4).get_beatwise_sections(offset=0.25, beats=[Beat(1), Beat(1), Beat(1)]) == [[0.75, 2],
                                                                                                            1.25]
        assert QuarterDuration(4).get_beatwise_sections(offset=0.5,
                                                        beats=[Beat(1), Beat(1.5), Beat(1.5), Beat(0.5)]) == [
                   [0.5, 3, 0.5],
                   None]
        assert QuarterDuration(3).get_beatwise_sections(offset=0.5, beats=[Beat(1), Beat(1), Beat(0.5), Beat(0.5)]) == [
            [0.5, 2], 0.5]
        assert QuarterDuration(4).get_beatwise_sections(offset=0.15, beats=[Beat(0.5), Beat(1.5), Beat(1)]) == [
            [0.35, 1.5, 1], 1.15]

    def test_copy_quarter_duration(self):
        qd = QuarterDuration(3, 4)
        copied = qd.__copy__()
        assert id(qd.value) != id(copied.value)
        assert qd.value == copied.value

    def test_check_quarter_duration(self):
        assert _check_quarter_duration_value(1)
        assert _check_quarter_duration_value(1.2)
        assert _check_quarter_duration_value(1 / 3)
        assert _check_quarter_duration_value(Fraction(1, 3))
        assert _check_quarter_duration_value(QuarterDuration(1, 3))

    def test_get_type_and_dots(self):
        assert QuarterDuration(1)._get_type_and_dots() == ('quarter', 0)
        assert QuarterDuration(1.5)._get_type_and_dots() == ('quarter', 1)
        assert QuarterDuration(1.75)._get_type_and_dots() == ('quarter', 2)

        assert QuarterDuration(0.75)._get_type_and_dots() == ('eighth', 1)
        assert QuarterDuration(2 + 2 / 2 + 2 / 4)._get_type_and_dots() == ('half', 2)

        assert QuarterDuration(1 / 3)._get_type_and_dots() == ('eighth', 0)
        assert QuarterDuration(2 / 3)._get_type_and_dots() == ('quarter', 0)

    def test_get_type(self):
        assert QuarterDuration(1).get_type() == 'quarter'
        assert QuarterDuration(1.5).get_type() == 'quarter'
        assert QuarterDuration(1.75).get_type() == 'quarter'

        assert QuarterDuration(0.75).get_type() == 'eighth'
        assert QuarterDuration(2 + 2 / 2 + 2 / 4).get_type() == 'half'

        assert QuarterDuration(1 / 3).get_type() == 'eighth'
        assert QuarterDuration(2 / 3).get_type() == 'quarter'

    def test_get_dots(self):
        assert QuarterDuration(1).get_number_of_dots() == 0
        assert QuarterDuration(1.5).get_number_of_dots() == 1
        assert QuarterDuration(1.75).get_number_of_dots() == 2

        assert QuarterDuration(0.75).get_number_of_dots() == 1
        assert QuarterDuration(2 + 2 / 2 + 2 / 4).get_number_of_dots() == 2

    def test_is_not_writable(self):
        for value in [0.23, 5 / 2, 9 / 2, 5 / 4, 9 / 5]:
            with self.assertRaises(QuarterDurationIsNotWritable):
                QuarterDuration(value)._get_type_and_dots()
            with self.assertRaises(QuarterDurationIsNotWritable):
                QuarterDuration(value).get_type()
            with self.assertRaises(QuarterDurationIsNotWritable):
                QuarterDuration(value).get_number_of_dots()
